import numpy as np

import reseval
from .base import Base


###############################################################################
# Word selection test
###############################################################################


class WordSelect(Base):

    def __init__(self, config, directory):
        super().__init__(config, directory)

        # Filter text files
        self.files = [
            file for file in self.files
            if not file.name.endswith('-words.txt')]

    @classmethod
    def analyze(cls, conditions, responses, random_seed=0):
        """Perform statistical analysis on evaluation results"""
        # Get number of times each stem is annotated
        stem_counts = {
            stem: len(response) for stem, response in responses.items()}

        # Get total number of annotations
        results = {
            'samples': sum(stem_counts.values()),
            'scores': cls.parse(responses) }

        return results, stem_counts

    @classmethod
    def parse(cls, responses):
        """Parse responses to produce per-word emphasis scores"""
        results = {}
        for stem, scores in responses.items():
            scores = np.array([
                [float(character) for character in score] for score in scores])
            results[stem] = scores.mean(axis=0).tolist()
        return results

    def response_type(self):
        """Retrieve the MySQL datatype of a participant response"""
        # For WordSelect, we store an array of booleans indicating whether each
        # word was selected. We assume a maximum of 255 words.
        return 'varchar(255)'

    def validate(self):
        """Perform validation on the directory of files to evaluate"""
        super().validate()

        # Iterate over files
        for stem, _ in self.stems_and_extensions():

            # Ensure corresponding text file is present
            if not (self.directory / f'{stem}-words.txt').exists():
                raise reseval.ValidationError()



import contextlib
import csv
import os

import mysql.connector

import reseval


###############################################################################
# Constants
###############################################################################


# MySQL database schema
SCHEMA = [
    # Evaluation conditions. For example, for an AB test, these are the
    # names of the conditions corresponding to the "A" and "B" choices.
    #
    # Condition - The name of the condition
    'CREATE TABLE `conditions` ('
    '  `Condition` varchar(255) UNIQUE,'
    '  PRIMARY KEY (`Condition`)'
    ')',

    # Tables of files to be evaluated, split into stem and extension.
    # Exact file paths depend on the test being performed. See README for
    # details.
    #
    # Stem - The file stem
    # Extension - The file extension
    # Uses - The number of times this stem has been selected for evaluation
    'CREATE TABLE `files` ('
    '  `Stem` varchar(255) UNIQUE,'
    '  `Extension` varchar(16) NOT NULL,'
    '  `Uses` int DEFAULT 0,'
    '  PRIMARY KEY (`Stem`)'
    ')',

    # Participant information
    # This assumes each participant only takes the survey once.
    # Prescreening questions from the configuration file are included
    # as well. These questions are formatted as
    # f' `{name}` varchar({length}) NOT NULL,', where "name" is the
    # name given to the question in the configuration file, and "length"
    # is automatically computed from the longest allowed answer.
    #
    # ID - The participant ID
    # CompletionCode - The random code the participant recieves upon completion
    # DateTaken - The time at which the participant took the survey
    'CREATE TABLE `participants` ('
    '  `ID` char(32) UNIQUE,'
    '{}'
    '  `CompletionCode` char(10) DEFAULT \'INCOMPLETE\','
    '  `DateTaken` timestamp DEFAULT CURRENT_TIMESTAMP,'
    '  PRIMARY KEY (`ID`)'
    ')',

    # Participant's responses
    # The format of the response depends on the test being taken. For example,
    # an ABX test has as responses a character A, B, or X, and an MOS test has
    # as responses numbers 1-5.
    #
    # Stem - The file stem
    # Participant - The participant ID
    # OrderAsked - The chronological ordering of the survey
    # Response - The response given by the participant
    'CREATE TABLE `responses` ('
    '  `Stem` varchar(255),'
    '  `Participant` char(32),'
    '  `OrderAsked` int,'
    '  `Response` {},'
    '  FOREIGN KEY (`Stem`) REFERENCES `files`(`Stem`),'
    '  FOREIGN KEY (`Participant`) REFERENCES `participants`(`ID`)'
    ')']

# Names of tables in the database
TABLES = ['conditions', 'files', 'participants', 'responses']


###############################################################################
# Database management
###############################################################################


def create(config, test, local=False):
    """Write database environment variable file and initialize the database"""
    print('Creating database...')

    # Create new database and retrieve credentials
    credentials = module(config, local).create(config['name'])

    # Save environment variables
    for key, value in credentials.items():
        os.environ[key] = value

    # Connect to MySQL database
    with connect() as (_, cursor):

        # Create database tables
        for table, command in zip(TABLES, SCHEMA):

            # Add prescreen questions from configuration file
            if table == 'participants':
                questions = ''

                all_questions = []
                if 'prescreen_questions' in config:
                    all_questions += config['prescreen_questions']
                if 'followup_questions' in config:
                    all_questions += config['followup_questions']
                for question in all_questions:
                    name = question['name']

                    # Get maximum answer length
                    question_type = question['type']
                    if question_type == 'free-response':
                        max_length = 1024
                    elif question_type == 'multiple-choice':
                        answers = question['answers']
                        max_length = max([len(answer) for answer in answers])
                    else:
                        raise ValueError(
                            f'Question type {question_type} is not recognized')

                    # Add database row
                    questions += f'  `{name}` varchar({max_length}),'

                # Update MySQL command with prescreen questions
                command = command.format(questions)

            # Response type is unique to the test
            if table == 'responses':
                command = command.format(test.response_type())

            # Communicate with database
            cursor.execute(command)

    # Upload conditions and filenames
    upload_test(test)

    # Upload any previous results from this evaluation
    upload_previous(config['name'])

    # Write credentials to environment file
    environment_file = (
        reseval.EVALUATION_DIRECTORY /
        config['name'] /
        'credentials' /
        '.env')
    environment_file.parent.mkdir(exist_ok=True, parents=True)
    with open(environment_file, 'w') as file:
        for key, value in credentials.items():
            file.write(f'{key}={value}\n')

    return credentials


def destroy(name):
    """Destroy a database"""
    # Load database credentials
    try:
        reseval.load.environment_variables_by_name(name)
    except FileNotFoundError:
        return

    # Destroy database
    local = reseval.is_local(name)
    config = reseval.load.config_by_name(name)
    module(config, local).destroy(name)

    # Cleanup credentials
    (
        reseval.EVALUATION_DIRECTORY /
        name /
        'credentials' /
        '.env'
    ).unlink(missing_ok=True)


def download(name, directory, tables=TABLES):
    """Download the contents of the MySQL database"""
    # Load database credentials
    reseval.load.environment_variables_by_name(name)

    # Create directory to store results
    directory.mkdir(exist_ok=True, parents=True)

    with connect() as (_, cursor):

        # Download each table to a csv file
        for name in tables:

            # Download table
            cursor.execute('SELECT * FROM ' + name)
            table = cursor.fetchall()

            # Get table header
            cursor.execute(f'SHOW COLUMNS FROM {name}')
            header = [column[0] for column in cursor.fetchall()]

            # Write table to CSV
            with open(directory / f'{name}.csv', 'w') as file:
                file = csv.writer(file)
                file.writerow(header)
                file.writerows(table)


###############################################################################
# Utilities
###############################################################################


@contextlib.contextmanager
def connect():
    """Connect to a MySQL database"""
    try:

        # Connect to the database
        connection = mysql.connector.connect(
            database=os.environ['MYSQL_DBNAME'],
            host=os.environ['MYSQL_HOST'],
            user=os.environ['MYSQL_USER'],
            password=os.environ['MYSQL_PASS'])

        # Create cursor to execute commands
        cursor = connection.cursor()

        # Execute user code
        yield connection, cursor

    finally:

        # Close database connection
        if 'cursor' in locals():
            cursor.close()
        if 'connection' in locals():
            connection.close()


def upload_previous(name):
    """Upload previous results from this evaluation to database"""
    # Check if we have previous results to upload
    tables = ['participants', 'responses']
    directory = reseval.EVALUATION_DIRECTORY / name / 'tables'
    if not all((directory / f'{table}.csv').exists() for table in tables):
        return

    # Connect to MySQL database
    with connect() as (connection, cursor):

        # Upload previous results
        for table in tables:

            # Read CSV
            with open(directory / f'{table}.csv') as file:
                rows = [row for row in csv.DictReader(file)]

            # Skip if table is empty
            if len(rows) == 0:
                return

            # Format MySQL command
            keys = list(rows[0].keys())
            columns, values, update = '', '', ''
            for i, key in enumerate(keys):
                columns += f'`{key}`'
                values += '%s'
                update += f'`{key}`=`{key}`'
                if i != len(keys) - 1:
                    columns += ', '
                    values += ', '
                    update += ', '
            command = (
                f'INSERT INTO {table} ({columns}) VALUES ({values}) '
                f'ON DUPLICATE KEY UPDATE {update}')

            # Specify data order
            items = [tuple(row[key] for key in keys) for row in rows]

            # Execute insertions
            cursor.executemany(command, items)

        # Communicate with database
        connection.commit()


def upload_test(test):
    """Upload test information to database"""
    # Connect to MySQL database
    with connect() as (connection, cursor):

        # Add conditions to the database
        command = (
            'INSERT INTO conditions (`Condition`) VALUES (%s) '
            'ON DUPLICATE KEY UPDATE `Condition`=`Condition`')
        cursor.executemany(command, [(c,) for c in test.conditions])

        # Add filenames and extensions to database
        command = (
            'INSERT INTO files (Stem, Extension) VALUES (%s, %s) ' +
            'ON DUPLICATE KEY UPDATE Stem=Stem, Extension=Extension')

        stems_and_extensions = test.stems_and_extensions()
        cursor.executemany(command, stems_and_extensions)
        connection.commit()


def module(config, local=False):
    """Get the database module to use"""
    if local:
        return reseval.database.localhost
    database = config['database']
    if database == 'aws':
        return reseval.database.aws
    if database == 'heroku':
        return reseval.database.heroku
    raise ValueError(f'Database service {database} is not recognized')



import os
import random
from abc import ABC, abstractmethod

import reseval


###############################################################################
# Abstract base test class
###############################################################################


class Base(ABC):

    def __init__(self, config, directory):
        self.directory = directory
        self.random_seed = config['random_seed']
        self.participants = config['participants']
        self.samples_per_participant = config['samples_per_participant']

        # Get all files in directory relative to the directory
        self.files = [
            file.relative_to(directory)
            for file in directory.rglob('*') if file.is_file()]

        # Get the set of conditions
        self.conditions = sorted([
            path.name for path in directory.iterdir() if path.is_dir()])

    @classmethod
    @abstractmethod
    def analyze(cls, conditions, responses, random_seed=0):
        """Perform statistical analysis on evaluation results"""
        pass

    def assign(self, random_seed=0):
        """Randomly assign files to each participant"""
        # Seed random number generation
        random.seed(random_seed)

        # Get shuffled files
        files = [
            stem.replace("\\", "/") + extension
            for stem, extension in self.stems_and_extensions()]
        random.shuffle(files)

        # Assign stems to participants
        index = 0
        samples = self.samples_per_participant
        assignments, residual = [], []

        # We generate more assignments than the expected number of participants
        # in case participants leave during the test or we extend the test
        while len(assignments) < 10 * self.participants:

            # Shuffle and reset index whenever we reach the end
            while index + samples - len(residual) >= len(files):
                residual.extend(files[index:])
                random.shuffle(files)
                index = 0

            # Add assignment
            end = index + samples - len(residual)
            assignments.append(residual + files[index:end])
            index = end

            # Reset residual
            residual = []

        return assignments

    @classmethod
    def plot(cls, results, file):
        """Create a plot of the results and save to disk"""
        pass

    def response_type(self):
        """Retrieve the MySQL datatype of a participant response"""
        # By default, the response type is the name of the winning condition.
        max_length = max([len(condition) for condition in self.conditions])
        return f'varchar({max_length})'

    def stems(self):
        """Retrieve the file stems to be uploaded to the database"""
        return [stem for stem, _ in self.stems_and_extensions()]

    def stems_and_extensions(self):
        """Retrieve file stems and extensions to be updated to the database"""
        if self.conditions:
            cond = sorted([
                file for file in self.files
                if file.parts[0] == self.conditions[0]])
            return [os.path.splitext('/'.join(file.parts[1:])) for file in cond]
        return [(file.stem, file.suffix) for file in self.files]

    def validate(self):
        """Perform validation on the directory of files to evaluate"""
        # Use first condition as reference
        common = sorted([
            file for file in self.files
            if file.parts[0] == self.conditions[0]])

        # All subsequent conditions must match
        for condition in self.conditions[1:]:
            files = sorted([
                file for file in self.files if file.parts[0] == condition])

            # Check lengths
            if len(files) != len(common):
                raise reseval.ValidationError()

            # Check file matches
            for a, b in zip(files, common):
                if a.parts[1:].join('/') != b.parts[1:].join('/'):
                    raise reseval.ValidationError()



import itertools
import random

import reseval
from .base import Base


###############################################################################
# MOS test
###############################################################################


class MOS(Base):

    @classmethod
    def analyze(cls, conditions, responses, random_seed=0):
        """Perform statistical analysis on evaluation results"""
        # Get MOS ratings
        condition_scores, stem_scores = cls.scores(conditions, responses)

        # Iterate over unique pairs of conditions
        samples = sum(len(scores) for scores in condition_scores.values())
        all_results = {'samples': samples, 'conditions': {}}
        for condition_a, condition_b in itertools.combinations(conditions, 2):
            results = {}

            # Two-sided t-test assuming equal variance
            results['student_ttest'] = reseval.stats.student_ttest(
                condition_scores[condition_a],
                condition_scores[condition_b],
                random_seed)

            # Two-sided t-test assuming unequal variance
            results['welchs_ttest'] = reseval.stats.welchs_ttest(
                condition_scores[condition_a],
                condition_scores[condition_b],
                random_seed)

            # Two-sided Wilcoxon test with no distribution assumptions
            results['mannwhitney'] = reseval.stats.mannwhitney(
                condition_scores[condition_a],
                condition_scores[condition_b])

            all_results['conditions'][f'{condition_a}-{condition_b}'] = results

        return all_results, stem_scores

    @classmethod
    def plot(cls, results, file):
        """Create a plot of the results and save to disk"""
        reseval.plot.violin(results, file, range(0, 6))

    def response_type(self):
        """Retrieve the MySQL datatype of a participant response"""
        # For MOS, we store the condition of the file presented to the
        # participant as well as the score given to the file by the participant
        # (an integer 1-5 inclusive). The format is "<condition>-<score>".
        max_length = max([len(condition) for condition in self.conditions])
        return f'varchar({max_length + 2})'

    @classmethod
    def scores(cls, conditions, responses):
        """Get scores from participant responses"""
        condition_scores = {cond: [] for cond in conditions}
        stem_scores = {
            stem: {cond: [] for cond in conditions} for stem in responses}
        for stem, results in responses.items():
            for result in results:
                parts = result.split('-')
                condition, score = '-'.join(parts[:-1]), int(parts[-1])
                condition_scores[condition].append(score)
                stem_scores[stem][condition].append(score)
        return condition_scores, stem_scores

    def assign_conditions(self, random_seed=0):
        """Randomly assign conditions to each participant"""
        # Seed random number generation
        random.seed(random_seed)

        # Get shuffled conditions
        all_conditions = self.conditions
        random.shuffle(all_conditions)

        # Assign conditions to participants
        index = 0
        samples = self.samples_per_participant
        assigned_conditions, residual = [], []

        # We generate more conditions than the expected number of participants
        # in case participants leave during the test or we extend the test
        while len(assigned_conditions) < 10 * self.participants:

            # Shuffle and reset index whenever we reach the end
            while index + samples - len(residual) >= len(all_conditions):
                residual.extend(all_conditions[index:])
                random.shuffle(all_conditions)
                index = 0

            # Add conditions
            end = index + samples - len(residual)
            assigned_conditions.append(residual + all_conditions[index:end])
            index = end

            # Reset residual
            residual = []

        return assigned_conditions



import reseval
from .base import Base


###############################################################################
# AB test
###############################################################################


class AB(Base):

    @classmethod
    def analyze(cls, conditions, responses, random_seed=0):
        """Perform statistical analysis on evaluation results"""
        # Get counts of responses for each condition and for each stem
        condition_counts, stem_counts = cls.counts(conditions, responses)

        # Two-sided Binomial test
        results = {
            'binomial': reseval.stats.binomial(
                [condition_counts[condition] for condition in conditions])}

        # Format results
        results = {
            'samples': sum(condition_counts.values()),
            'conditions': {f'{conditions[0]}-{conditions[1]}': results}}

        return results, stem_counts

    @classmethod
    def counts(cls, conditions, responses):
        """Get counts from participant responses"""
        condition_counts = {cond: 0 for cond in conditions}
        stem_counts = {
            stem: {cond: 0 for cond in conditions} for stem in responses}
        for stem, results in responses.items():
            for result in results:
                condition_counts[result] += 1
                stem_counts[stem][result] += 1
        return condition_counts, stem_counts

    @classmethod
    def plot(cls, results, file):
        """Create a plot of the results and save to disk"""
        reseval.plot.barh(results, file)



import json
import os
import typing
from pathlib import Path

import reseval


###############################################################################
# Get subjective evaluation results
###############################################################################


def results(
    name: str,
    directory: typing.Union[str, bytes, os.PathLike] = Path()) -> dict:
    """Get the results of a subjective evaluation

    Args:
        name: The name of the subjective evaluation to retrieve results for
        directory: The directory to save results

    Returns:
        dict: Evaluation results
    """
    # Download crowdsource results
    crowdsource = reseval.crowdsource.assignments(name)

    # Download database tables
    reseval.database.download(
        name,
        reseval.EVALUATION_DIRECTORY / name / 'tables')
    if directory is not None:
        reseval.database.download(name, directory / name / 'tables')

    # Load responses
    config = reseval.load.config_by_name(name)
    conditions = reseval.load.conditions(name)
    responses = reseval.load.responses(name)
    participants = reseval.load.participants(name)

    # Match participant IDs
    for assignment in crowdsource:
        try:
            completion_code = assignment['Answer'].split(
                '<FreeText>')[1].split('</FreeText>')[0].replace('&amp;', '&')
            participant = [
                p['ID'] for p in participants
                if p['CompletionCode'] == completion_code]
            assignment['ParticipantID'] = participant[0] if participant else ''
        except IndexError as error:
            assignment['ParticipantID'] = ''

    # Save crowdsource results
    crowdsource_file = directory / name / 'crowdsource' / 'crowdsource.json'
    crowdsource_file.parent.mkdir(exist_ok=True, parents=True)
    with open(crowdsource_file, 'w') as file:
        json.dump(crowdsource, file, indent=4, default=str)

    # No responses yet
    if len(responses) == 0:
        results = {'samples': 0, 'conditions': {}}

        # Save results
        with open(directory / name / 'results.json', 'w') as file:
            json.dump(results, file, indent=4)

        return results

    # Get condition names
    conditions = [condition['Condition'] for condition in conditions]

    # TEMPORARY
    # responses, residual = reseval.filter.anchors(
    #     responses,
    #     conditions,
    #     'bitcrush',
    #     'original')

    # TEMPORARY - filter out incomplete evaluations
    # responses = reseval.filter.count(
    #     responses,
    #     config['samples_per_participant'])

    # Group results by file stems
    responses_by_stem = {}
    for response in responses:
        stem = response['Stem']
        if stem in responses_by_stem:
            responses_by_stem[stem].append(response['Response'])
        else:
            responses_by_stem[stem] = [response['Response']]

    # Get test
    test = reseval.test.get(config)

    # Analyze results
    analysis, stem_scores = test.analyze(
        conditions,
        responses_by_stem,
        config['random_seed'])

    # Save results
    with open(directory / name / 'results.json', 'w') as file:
        json.dump(analysis | {'stems': stem_scores}, file, indent=4)

    # Create and save figure
    test.plot(stem_scores, directory / name / 'results.pdf')

    return analysis



import argparse
from pathlib import Path

import reseval


###############################################################################
# Entry point
###############################################################################


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(
        description='Create a subjective evaluation')
    parser.add_argument('config', type=Path, help='The configuration file')
    parser.add_argument(
        'directory',
        type=Path,
        help='The directory containing the files to evaluate')
    parser.add_argument(
        '--local',
        action='store_true',
        help='Run subjective evaluation locally')
    parser.add_argument(
        '--production',
        action='store_true',
        help='Deploy the subjective evaluation to crowdsource participants')
    return parser.parse_args()


reseval.create(**vars(parse_args()))



import math
import scipy.stats


###############################################################################
# Statistical tests
###############################################################################


def binomial(counts):
    """Performs a two-sided Binomial t-test"""
    result = scipy.stats.binomtest(counts[0], sum(counts))
    return {
        'Sample mean': result.proportion_estimate,
        'p-value': result.pvalue}


def student_ttest(x, y, random_seed=0):
    """Performs a two-sided Student t-test"""
    result = scipy.stats.ttest_ind(x, y, random_state=random_seed)
    return {
        'Test statistic': result.statistic,
        'p-value': result.pvalue}


def mannwhitney(x, y):
    """Performs a two-sided Mann-Whitney test"""
    # Handle no data
    if len(x) == 0 or len(y) == 0:
        return {'Test statistic': math.nan, 'p-value': math.nan}

    # Perform test
    result = scipy.stats.mannwhitneyu(x, y)

    return {
        'Test statistic': result.statistic,
        'p-value': result.pvalue}


def welchs_ttest(x, y, random_seed=0):
    """Performs a two-sided Welch's t-test"""
    result = scipy.stats.ttest_ind(
        x,
        y,
        equal_var=False,
        random_state=random_seed)
    return {
        'Test statistic': result.statistic,
        'p-value': result.pvalue}


def wilcoxon(x, y):
    """Performs a two-sided Wilcoxon signed-rank test"""
    # Handle no data
    if len(x) == 0:
        return

    # Perform test
    result = scipy.stats.wilcoxon(x, y)

    return {
        'Test statistic': result.statistic,
        'p-value': result.pvalue}



import reseval
from .base import Base


###############################################################################
# ABX test
###############################################################################


class ABX(Base):

    @classmethod
    def analyze(cls, conditions, responses, random_seed=0):
        """Perform statistical analysis on evaluation results"""
        conditions = [cond for cond in conditions if cond != 'reference']
        return reseval.test.AB.analyze(conditions, responses, random_seed)

    @classmethod
    def plot(cls, results, file):
        """Create a plot of the results and save to disk"""
        reseval.plot.barh(results, file)



import argparse

import reseval


###############################################################################
# Entry point for destroying subjective evaluation compute resources
###############################################################################


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(
        description='Clean-up database, server, and storage resources')
    parser.add_argument('name', help='The name of the evaluation to destroy')
    parser.add_argument(
        '--force',
        action='store_true',
        help='Destroy evaluation resources even if it is still running. ' +
             'Pays all participants who have taken the evaluation.')
    return parser.parse_args()


reseval.destroy(**vars(parse_args()))



import argparse

import reseval


###############################################################################
# Entry point for monitoring subjective evaluations
###############################################################################


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(
        description='Monitor subjective evaluations')
    parser.add_argument(
        '--names',
        nargs='+',
        help='The names of the evaluations to monitor. ' +
             'If not provided, monitors all evaluations.')
    parser.add_argument(
        '--interval',
        type=int,
        default=60,
        help='The time between monitoring updates in seconds')
    return parser.parse_args()


reseval.monitor(**vars(parse_args()))



import argparse
from pathlib import Path

import reseval


###############################################################################
# Entry point for downloading the database tables of a subjective evaluation
###############################################################################


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(
        description='Download the database tables of a subjective evaluation')
    parser.add_argument('name', help='The name of the evaluation')
    parser.add_argument(
        'directory',
        default=Path(),
        help='The directory to populate with the database contents')
    parser.add_argument(
        '--tables',
        default=reseval.database.TABLES,
        help='The tables to download')
    return parser.parse_args()


reseval.database.download(**vars(parse_args()))



import re

import reseval


def extend(
    name: str,
    participants: int,
    local: bool = False,
    production: bool = False):
    """Extend a subjective evaluation

    Args:
        name: The name of the evaluation to extend
        participants: The number of new participants to add to the evaluation
        local: Run subjective evaluation locally
        production: Deploy the subjective evaluation to crowdsource
            participants
    """
    if not (reseval.EVALUATION_DIRECTORY / name).exists():
        raise ValueError(f'Cannot extend non-existant evaluation {name}')

    if production and local:
        raise ValueError('Cannot deploy production build locally')

    # We can only restart evaluations that have been run in the same mode
    prod_file = reseval.EVALUATION_DIRECTORY / name / '.prod'
    if not production and prod_file.exists():
        raise ValueError(
            f'Cannot extend production evaluation {name} '
            'in non-production mode')
    local_file = reseval.EVALUATION_DIRECTORY / name / '.local'
    if not local and local_file.exists():
        raise ValueError(
            f'Cannot extend local evaluation {name} '
            'in non-local mode')

    # Extend crowdsourcing task
    reseval.crowdsource.extend(name, participants)

    # Get saved configuration
    config_file = reseval.EVALUATION_DIRECTORY / name / 'config.yaml'
    with open(config_file, encoding='utf-8') as file:
        lines = file.readlines()

    # Replace number of participants
    for i in range(len(lines)):
        if lines[i].startswith('participants: '):
            previous = re.findall(r'\d+', lines[i])[0]
            current = int(previous) + participants
            lines[i] = re.sub(r'\d+', str(current), lines[i])

    # Save new configuration
    with open(config_file, 'w', encoding='utf-8') as file:
        for line in lines:
            file.write(line)



def anchors(responses, conditions, low_anchor, high_anchor):
    """Removes participants that rate the low anchor above the high anchor"""
    low_index = conditions.index(low_anchor)
    high_index = conditions.index(high_anchor)

    # Make blacklist of bad participants
    blacklist = set()
    for response in responses:
        scores = response['Response']
        scores = [scores[i:i + 3] for i in range(0, len(scores), 3)]
        if scores[low_index] > scores[high_index]:
            blacklist.add(response['Participant'])

    # Split responses using blacklist
    filtered, residual = [], []
    for response in responses:
        (
            residual if response['Participant'] in blacklist else filtered
        ).append(response)

    # Print the impact of the filter
    participants = set(response['Participant'] for response in responses)
    print(
        f'Filtered out {len(blacklist)} of the {len(participants)} '
        'participants that ranked the low-anchor above the high-anchor')

    return filtered, residual


def count(responses, num):
    """Filter out participants that have an incorrect number of responses"""
    # Count responses
    participants = {}
    for response in responses:
        participant = response['Participant']
        if participant not in participants:
            participants[participant] = 1
        else:
            participants[participant] += 1

    # Filter out bad participants
    blacklist = [
        participant for participant, count in participants.items()
        if count != num]
    filtered = []
    for response in responses:
        if response['Participant'] not in blacklist:
            filtered.append(response)

    # Print the impact of the filter
    print(f'Filtered out {len(blacklist)} participants: {blacklist}')

    return filtered



import time
import typing

import rich
import rich.progress

import reseval


###############################################################################
# Monitor subjective evaluations
###############################################################################


def monitor(
    names: typing.Optional[typing.List[str]],
    interval: int = 60,
    exit_on_finish: bool = False):
    """Monitor subjective evaluations

    Args:
        names: The names of the evaluations to monitor. If not provided,
            monitors all evaluations.
        interval: The time between monitoring updates in seconds
        exit_on_finish: Whether to exit monitoring when the evaluation finishes
    """
    # List evaluations
    if names is None:
        names = [
            eval.name for eval in reseval.EVALUATION_DIRECTORY.glob('*')
            if eval.is_dir()]

    # Get total samples for each evaluation
    configs = [reseval.load.config_by_name(evaluation) for evaluation in names]
    totals = [
        cfg['participants'] * cfg['samples_per_participant']
        for cfg in configs]

    # Get total participants for each evaluation
    participants = [cfg['participants'] for cfg in configs]

    # Setup monitoring display
    analyses = [
        reseval.results(evaluation, reseval.EVALUATION_DIRECTORY)
        for evaluation in names]

    # Render display and monitor
    content = displays(names, totals, participants, analyses)
    with rich.progress.Live(content, refresh_per_second=.2) as live:

        # Monitor evaluations
        while True:

            # Iterate over evaluations
            for index, (evaluation, total) in enumerate(zip(names, totals)):

                # Get current progress
                reseval.database.download(
                    evaluation,
                    reseval.EVALUATION_DIRECTORY / evaluation / 'tables')
                count = len(reseval.load.responses(evaluation))

                # Update display if we have new results
                if count != analyses[index]['samples']:

                    # Get current statistics
                    analysis = reseval.results(
                        evaluation,
                        reseval.EVALUATION_DIRECTORY)
                    analyses[index] = analysis

                    # Update display
                    live.update(displays(
                        names,
                        totals,
                        participants,
                        analyses))

            # Maybe exit when the evaluations are done
            if exit_on_finish:

                done = True
                for name, analysis, total in zip(names, analyses, totals):

                    # Don't exit if the evaluation is still active
                    if (
                        (reseval.is_local(name) and analysis['samples'] < total) or
                        reseval.crowdsource.active(name)
                    ):
                       done = False

                if done:
                    break

            # Wait a while
            time.sleep(interval)


def display(name, total, participants, analysis):
    """Format one evaluation for display"""
    grid = rich.progress.Table.grid()
    grid.add_column()

    # Get current number of participants
    if reseval.is_local(name):
        current = len(reseval.load.participants(name))
    else:
        current = reseval.crowdsource.progress(name)

    # Format a table of statistics
    table = rich.progress.Table.grid(padding=(0, 2))
    table.add_column(justify='right')
    table.add_column(justify='center')
    table.add_column(justify='center')
    table.add_column(justify='right')
    table.add_row('participants', f'{current}/{participants}')
    table.add_row('samples', f'{analysis["samples"]}/{total}')

    if 'conditions' in analysis:
        for condition, items in analysis['conditions'].items():
            for i, (test, values) in enumerate(items.items()):
                for j, (key, val) in enumerate(values.items()):
                    table.add_row(
                        condition if i == 0 and j == 0 else None,
                        test if j == 0 else None,
                        key,
                        str(val))

    # Create progress bar
    bar = rich.progress.Progress(
        rich.progress.BarColumn(),
        rich.progress.TextColumn(
            '[progress.percentage]{task.percentage:>3.0f}%'))

    # Add task to progress bar
    bar.add_task(name, total=total, completed=analysis['samples'])

    # Add table and bar to grid
    grid.add_row(bar)
    grid.add_row(table)

    # return table, bar
    return rich.panel.Panel(grid, title=name)


def displays(names, totals, participants, analyses):
    """Format multiple evaluations for display"""
    displays = rich.progress.Table.grid()
    iterator = zip(names, totals, participants, analyses)
    for name, total, participant, analysis in iterator:
        displays.add_row(display(name, total, participant, analysis))
    return displays


def format_stats(stats):
    """Format statistics for printing to a table"""
    result = ''
    for key, value in stats.items():
        result += f'{key} - {value}\n'
    return result[:-1]



import shutil
import reseval


###############################################################################
# Cleanup subjective evaluation compute resources
###############################################################################


def destroy(name: str, force: bool = False):
    """Destroys an evaluation's storage, database, server, and crowdsource task

    Args:
        name: The name of the evaluation to destroy
        force: Destroy evaluation resources even if it is still running. Pays
            all participants who have taken the evaluation.
    """
    # Evaluation doesn't exist
    if not (reseval.EVALUATION_DIRECTORY / name).exists():
        return

    # Is the evaluation currently active?
    try:
        active = reseval.crowdsource.active(name)
    except Exception:
        if force:
            active = False

    paid = reseval.crowdsource.paid(name)
    if active or not paid:

        if not force:
            tag = (
                ('active ' if active else '') +
                ('unpaid ' if not paid else ''))
            raise ValueError(
                f'Not destroying {tag}subjective evaluation {name}')

        else:

            # Pay participants
            if not paid:
                reseval.crowdsource.pay(name)

    # Maybe destroy crowdsource task
    reseval.crowdsource.destroy(name)

    # Destroy database
    reseval.database.destroy(name)

    # Destroy server
    reseval.server.destroy(name)

    # Destroy storage
    reseval.storage.destroy(name)

    # Clean-up cache
    shutil.rmtree(reseval.EVALUATION_DIRECTORY / name)



import matplotlib
import matplotlib.pyplot as plt
import numpy as np


###############################################################################
# Statistical plots
###############################################################################


def barh(results, file):
    """Create a barh plot"""
    # TODO
    pass


def violin(results, file, yticks):
    """Create a violin plot"""
    # Set font size
    matplotlib.rcParams.update({'font.size': 18})

    # Organize data by condition
    data = {}
    for _, values in results.items():
        for condition, scores in values.items():
            if condition not in data:
                data[condition] = scores
            else:
                data[condition].extend(scores)
    keys = np.array(list(data.keys()))
    values = np.array(
        [np.array(sorted(value), dtype=int) for value in data.values()])

    # Order data by average value
    means = np.nan_to_num([value.mean() for value in values])

    # TEMPORARY
    # print([(key, mean) for key, mean in zip(keys, means)])

    # Get medians
    _, medians, _ = np.percentile(
        values,
        [25, 50, 75],
        axis=values.ndim - 1)

    indices = np.argsort(means)
    means, keys, values, medians = (
        means[indices],
        list(keys[indices]),
        values[indices],
        medians[indices])

    # No data
    if not all(len(value) for value in values):
        return

    # Create violin plot
    figure, ax = plt.subplots(1, 1, squeeze=True, figsize=(16, 6))
    ax.violinplot(
        values.T,
        showmeans=False,
        showmedians=False,
        showextrema=False)

    # Add mean and median markers
    inds = np.arange(1, len(medians) + 1)
    ax.scatter(inds, means, marker='o', color='black', s=80, zorder=3)
    ax.scatter(inds, medians, marker='o', color='white', s=80, zorder=3)

    # Style axes
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['left'].set_visible(False)
    ax.xaxis.set_ticks_position('none')
    ax.yaxis.set_ticks_position('none')
    for y in yticks:
        ax.hlines(
            y=y,
            xmin=0,
            xmax=len(keys) + .3,
            linewidth=1,
            color='#666666',
            alpha=.3)

    # TEMPORARY
    # label_map = {
    #     'bitcrush': 'Low anchor',
    #     'bottleneck': 'ASR bottleneck',
    #     'w2v2fc': 'Charsiu',
    #     'mel': 'Mels',
    #     'encodec': 'EnCodec',
    #     'w2v2fb': 'Wav2vec 2.0',
    #     'original': 'Original',
    # }

    # Add labels
    ax.set_xticks(
        np.arange(1, len(keys) + 1),
        # labels=[label_map[key] for key in keys],
        labels=keys,
        fontsize=16)
    ax.set_xlim(0.25, len(keys) + 0.75)

    # Save figure
    figure.savefig(file, bbox_inches='tight', pad_inches=0)


###############################################################################
# Utilities
###############################################################################


def adjacent_values(vals, q1, q3):
    upper_adjacent_value = q3 + (q3 - q1) * 1.5
    upper_adjacent_value = np.clip(upper_adjacent_value, q3, vals[-1])
    lower_adjacent_value = q1 - (q3 - q1) * 1.5
    lower_adjacent_value = np.clip(lower_adjacent_value, vals[0], q1)
    return lower_adjacent_value, upper_adjacent_value



import http.client
import json
import os
import tarfile
import tempfile
import time
from pathlib import Path

import reseval


###############################################################################
# Heroku server management
###############################################################################


def create(name, detach=True):
    """Create a Heroku server"""
    # Get unique identifier
    unique = reseval.load.credentials_by_name(name, 'unique')['unique']

    # Connect to Heroku
    connection = http.client.HTTPSConnection('api.heroku.com')

    # Maybe install client
    client_directory = reseval.CACHE / 'client'
    if not (client_directory / 'node_modules').exists():
        with reseval.chdir(client_directory):
            reseval.npm.install().wait()

    # Build client
    with reseval.chdir(reseval.CACHE / 'client'):
        reseval.npm.build().wait()

    # Create a tarball of all files needed by the Heroku server
    with tempfile.TemporaryDirectory() as directory:
        tarball = Path(directory) / 'reseval.tar.gz'
        files = [
            'client/build',
            'server',
            'package-lock.json',
            'package.json',
            'Procfile',
            'server.ts',
            'tsconfig.json']
        with tarfile.open(tarball, 'w:gz') as tar:
            for file in files:
                tar.add(reseval.CACHE / file)

        # Upload tarball and get a URL
        tarball_url = reseval.storage.upload(name, tarball)

        # Server configuration
        data = {
            'source_blob': {'url': tarball_url},
            'buildpacks': [{
                'url': 'https://github.com/heroku/heroku-buildpack-nodejs',
                'name': 'heroku/nodejs'}]}
        headers = {
            'Accept': 'application/vnd.heroku+json; version=3',
            'Authorization': f'Bearer {os.environ["HerokuAccessKey"]}'}

        # Create server
        connection.request(
            'POST',
            f'/apps/{unique}/builds',
            json.dumps(data),
            headers=headers)

        # Get response from server
        response = json.loads(connection.getresponse().read().decode())

        # if app doesn't exist, raise error
        if response['id'] == 'not_found':
            raise ValueError(f'app name: {unique} does not exist')

    # Close the connection
    connection.close()

    # Wait until server is setup
    while status(name) == 'pending':
        time.sleep(3)

    if status(name) == 'failure':
        raise ValueError('Heroku server failed to start')

    # Return application URL
    return {'URL': f'http://{name}.herokuapp.com/'}


def destroy(name, credentials):
    """Destroy a Heroku server"""
    reseval.app.heroku.destroy(name)


###############################################################################
# Utilities
###############################################################################


def status(name):
    """Get current build status. One of ['succeeded', 'failed', 'pending']"""
    # Get unique identifier
    unique = reseval.load.credentials_by_name(name, 'unique')['unique']

    # Connect to Heroku
    connection = http.client.HTTPSConnection('api.heroku.com')

    # Send request
    reseval.load.api_keys()
    headers = {
        'Accept': 'application/vnd.heroku+json; version=3',
        'Authorization': f'Bearer {os.environ["HerokuAccessKey"]}'}
    connection.request('GET', f'/apps/{unique}/builds', headers=headers)

    # Get response
    data = json.loads(connection.getresponse().read().decode())

    # Close connection
    connection.close()

    # Get most recent build status from response
    status = list(map(lambda x: x['status'], data))
    return status[-1]



import contextlib
import os

import dotenv
import mysql.connector

import reseval


def create(name):
    """Create a local MySQL database"""
    # Get unique identifier
    unique = reseval.load.credentials_by_name(name, 'unique')['unique']

    # Load environment variables
    dotenv.load_dotenv(reseval.ENVIRONMENT_FILE)

    # Create connection
    with connect() as (_, cursor):

        # Create database
        cursor.execute(f'CREATE DATABASE `{unique}`')

        # Return credentials
        return {
            'MYSQL_DBNAME': unique,
            'MYSQL_HOST': os.environ['MYSQL_HOST'],
            'MYSQL_USER': os.environ['MYSQL_USER'],
            'MYSQL_PASS': os.environ['MYSQL_PASS']}


def destroy(name):
    """Destroy a local MySQL database"""
    # Get unique identifier
    unique = reseval.load.credentials_by_name(name, 'unique')['unique']

    # Create connection
    with connect() as (_, cursor):

        # Destroy database
        try:
            cursor.execute(f'DROP DATABASE `{unique}`')
        except mysql.connector.errors.DatabaseError:

            # Database doesn't exist
            pass


###############################################################################
# Constants
###############################################################################


@contextlib.contextmanager
def connect():
    """Connect to a local MySQL database"""
    try:

        # Connect to the database
        connection = mysql.connector.connect(
            host=os.environ['MYSQL_HOST'],
            user=os.environ['MYSQL_USER'],
            password=os.environ['MYSQL_PASS'])

        # Create cursor to execute commands
        cursor = connection.cursor()

        # Execute user code
        yield connection, cursor

    finally:

        # Close database connection
        if 'cursor' in locals():
            cursor.close()
        if 'connection' in locals():
            connection.close()



import json

import reseval


###############################################################################
# Server management
###############################################################################


def create(config, local=False, detach=False):
    """Create a server"""
    print('Creating server...')

    # Create server
    name = config['name']
    credentials = module(config, local).create(name, detach=detach)

    # Save server credentials
    file = reseval.EVALUATION_DIRECTORY / name / 'credentials' / 'server.json'
    file.parent.mkdir(exist_ok=True, parents=True)
    with open(file, 'w') as file:
        json.dump(credentials, file, indent=4, sort_keys=True)

    # Return the application URL
    return credentials['URL']


def destroy(name):
    """Destroy a server"""
    local = reseval.is_local(name)

    # Load config
    config = reseval.load.config_by_name(name)

    try:

        # Load credentials
        credentials = reseval.load.credentials_by_name(name, 'server')

    except FileNotFoundError:

        # We assume the server does not exist if we don't have credentials
        return

    # Destroy server
    module(config, local).destroy(name, credentials)

    # Cleanup credentials
    (
        reseval.EVALUATION_DIRECTORY /
        name /
        'credentials' /
        'server.json'
    ).unlink(missing_ok=True)


###############################################################################
# Utilities
###############################################################################


def module(config, local=False):
    """Get the crowdsourcing module to use"""
    if local:
        return reseval.server.local
    server = config['server']
    if server == 'aws':
        return reseval.server.aws
    if server == 'heroku':
        return reseval.server.heroku
    raise ValueError(f'Server hosting service {server} not recognized')



import itertools

import reseval
from .base import Base


###############################################################################
# MUSHRA test
###############################################################################


class MUSHRA(Base):

    @classmethod
    def analyze(cls, conditions, responses, random_seed=0):
        """Perform statistical analysis on evaluation results"""
        # Get MUSHRA ratings
        condition_scores, stem_scores = cls.scores(conditions, responses)

        # Iterate over unique pairs of conditions
        samples = len(condition_scores[conditions[0]])
        all_results = {'samples': samples, 'conditions': {}}
        for condition_a, condition_b in itertools.combinations(conditions, 2):
            results = {}

            # Two-sided t-test assuming equal variance
            results['student_ttest'] = reseval.stats.student_ttest(
                condition_scores[condition_a],
                condition_scores[condition_b],
                random_seed)

            # Two-sided t-test assuming unequal variance
            results['welchs_ttest'] = reseval.stats.welchs_ttest(
                condition_scores[condition_a],
                condition_scores[condition_b],
                random_seed)

            # Two-sided Wilcoxon test with no distribution assumptions
            results['wilcoxon'] = reseval.stats.wilcoxon(
                condition_scores[condition_a],
                condition_scores[condition_b])

            all_results['conditions'][f'{condition_a}-{condition_b}'] = results

        return all_results, stem_scores

    @classmethod
    def plot(cls, results, file):
        """Create a plot of the results and save to disk"""
        reseval.plot.violin(results, file, range(0, 101, 20))

    def response_type(self):
        """Retrieve the MySQL datatype of a participant response"""
        # For the MUSHRA test, the response is a string of concatenated
        # three-digit numbers, where each number is a value between 000 and
        # 100, inclusive. For example, if your MUSHRA test has four conditions
        # and the conditions, in alphabetical order, are given scores 67, 72,
        # 95, and 23, respectively, the response will be "067072095023".
        return f'varchar({3 * len(self.conditions)})'

    @classmethod
    def scores(cls, conditions, responses):
        """Get scores from participant responses"""
        condition_scores = {cond: [] for cond in conditions}
        stem_scores = {
            stem: {cond: [] for cond in conditions} for stem in responses}
        for stem, results in responses.items():
            for result in results:
                scores = [
                    int(result[i:i + 3]) for i in range(0, len(result), 3)]
                for condition, score in zip(sorted(conditions), scores):
                    condition_scores[condition].append(score)
                    stem_scores[stem][condition].append(score)
        return condition_scores, stem_scores



import subprocess


###############################################################################
# Npm server management

# NOTE - shell=True is often not considered best practice. This is for two
#     reasons:
#         1) It assumes the binary is the program you expect it to be. In
#            this case, npm.
#         2) If parameterized, it can run malicious code. For example,
#            f'npm run {user_input}' where
#            user_input = 'nonexistant_arg; rm -rf /'.
#     In our use case, we've asked users to install npm and we do not
#     parameterize the commands. The alternative is to set shell=False and
#     require the user to provide the path to the npm executable.
###############################################################################


def build(verbose=False):
    """Build the optimized front-end client code"""
    stdout = None if verbose else subprocess.DEVNULL
    return subprocess.Popen('npm run build', shell=True, stdout=stdout)


def install():
    """Installs an npm package in the current directory"""
    return subprocess.Popen('npm install', shell=True)


def start():
    """Start the server process"""
    return subprocess.Popen('npm run dev', shell=True)



import contextlib
import os
import typing
from pathlib import Path

import reseval


###############################################################################
# Run subjective evaluation pipeline
###############################################################################


def run(
    config: str,
    directory: typing.Union[str, bytes, os.PathLike],
    results_directory: typing.Union[str, bytes, os.PathLike] = Path(),
    local: bool = False,
    production: bool = False,
    interval: int = 120) -> dict:
    """Perform subjective evaluation

    Args:
        config: The configuration file
        directory: The directory containing the files to evaluate
        results_directory: The directory to save results
        local: Run subjective evaluation locally
        production: Deploy the subjective evaluation to crowdsource
            participants
        interval: The time between monitoring updates in seconds

    Returns:
        dict: Evaluation results
    """
    # Setup evaluation
    name = reseval.create(config, directory, local, production, detach=True)

    try:

        # Monitor evaluation until completion
        reseval.monitor([name], interval, exit_on_finish=True)

    except (Exception, KeyboardInterrupt) as exception:

        # Perform clean-up only if evaluation is local
        if local:
            print(exception)
        else:
            raise exception

    # Pay participants
    reseval.pay(name)

    # Get results
    results = reseval.results(name, results_directory)

    # Cleanup database, server, storage, and crowdsource task
    reseval.destroy(name, force=True)

    return results


###############################################################################
# Utilities
###############################################################################


@contextlib.contextmanager
def chdir(directory):
    """Change working directory"""
    curr_dir = os.getcwd()
    try:
        os.chdir(directory)
        yield
    finally:
        os.chdir(curr_dir)


def is_local(name):
    """Returns true if the evaluation is hosted locally"""
    return (reseval.EVALUATION_DIRECTORY / name / '.local').exists()



import json
import os
import requests
import string
import time

import boto3

import reseval


###############################################################################
# AWS RDS database management
###############################################################################


def create(name):
    """Create a MySQL database on AWS"""
    # Get unique identifier
    unique = reseval.load.credentials_by_name(name, 'unique')['unique']

    # Load API keys into environment variables
    reseval.load.api_keys()

    # Connect to EC2
    client = boto3.Session(
        aws_access_key_id=os.environ['AWSAccessKeyId'],
        aws_secret_access_key=os.environ['AWSSecretKey']
    ).client('ec2', region_name='us-east-1')

    # Get elastic beanstalk security groups
    security_groups = client.describe_security_groups(
        Filters=[
            {
                'Name': 'tag:elasticbeanstalk:environment-name',
                'Values': [unique]
            }
        ]
    )['SecurityGroups']

    # Add ingress rules for local IP and MySQL
    local_ip = json.loads(
        requests.get('https://api.seeip.org/jsonip?').text
    )['ip']
    group_ids = [
        group['GroupId'] for group in security_groups
        if any(tag['Value'] == 'AWSEBSecurityGroup' for tag in group['Tags'])]
    for group_id in group_ids:
        client.authorize_security_group_ingress(
            GroupId=group_id,
            IpPermissions=[
                {
                    'IpProtocol': 'tcp',
                    'IpRanges': [{'CidrIp': f'{local_ip}/32'}],
                    'FromPort': 22,
                    'ToPort': 22
                },
                {
                    'IpProtocol': 'tcp',
                    'IpRanges': [{'CidrIp': '0.0.0.0/0'}],
                    'FromPort': 3306,
                    'ToPort': 3306
                }
            ]
        )

    # Connect to AWS RDS
    client = connect()

    # Create random username and password so our database is harder to access
    random_user = (
        reseval.random.string(1, string.ascii_lowercase) +
        reseval.random.string(15))
    random_pass = (
        reseval.random.string(1, string.ascii_lowercase) +
        reseval.random.string(31))

    # Create database
    response = client.create_db_instance(
        DBName=unique,
        DBInstanceIdentifier=unique,
        DBInstanceClass='db.t2.micro',
        AllocatedStorage=64,
        Engine='mysql',
        MasterUsername=random_user,
        MasterUserPassword=random_pass,
        VpcSecurityGroupIds=group_ids,
        PubliclyAccessible=True)['DBInstance']

    # Get credentials
    while response['DBInstanceStatus'] == 'creating':
        time.sleep(3)
        response = client.describe_db_instances(
            DBInstanceIdentifier=unique
        )['DBInstances'][0]
    credentials = {
        'MYSQL_DBNAME': unique,
        'MYSQL_HOST': response['Endpoint']['Address'],
        'MYSQL_USER': random_user,
        'MYSQL_PASS': random_pass,
        'RDS_HOSTNAME': response['Endpoint']['Address'],
        'RDS_PORT': '3306',
        'RDS_DB_NAME': unique,
        'RDS_USERNAME': random_user,
        'RDS_PASSWORD': random_pass}

    return credentials


def destroy(name):
    """Destroy a MySQL database on AWS"""
    # Get unique identifier
    unique = reseval.load.credentials_by_name(name, 'unique')['unique']

    # Connect to AWS RDS
    client = connect()

    # Delete database
    client.delete_db_instance(
        DBInstanceIdentifier=unique,
        SkipFinalSnapshot=True,
        DeleteAutomatedBackups=True)


###############################################################################
# Utilities
###############################################################################


def connect():
    """Connect to AWS"""
    # Load API keys into environment variables
    reseval.load.api_keys()

    # Add credentials and connect
    return boto3.Session(
        aws_access_key_id=os.environ['AWSAccessKeyId'],
        aws_secret_access_key=os.environ['AWSSecretKey']
    ).client('rds', region_name='us-east-1')


def has_tag(group, tag):
    """Check if security group description contains a specified tag"""
    for tag in group['Tags']:
        if tag['Value'] == tag:
            return True
    return False



import argparse

import reseval


###############################################################################
# Entry point for paying participants
###############################################################################


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(
        description='Pay participants of a subjective evaluation')
    parser.add_argument('name', help='The name of the evaluation to pay for')
    return parser.parse_args()


reseval.pay(**vars(parse_args()))



import csv
import errno
import json
import os

import dotenv
import yaml

import reseval


###############################################################################
# File loading
###############################################################################


def api_keys():
    """Load cached API keys as environment variables"""
    with open(reseval.KEYS_FILE) as file:
        for line in file.readlines():
            key, value = line.rstrip().split('=', 1)
            os.environ[key] = value


def config_by_name(name):
    """Load configuration of an existing evaluation"""
    return config_from_file(reseval.EVALUATION_DIRECTORY / name / 'config.yaml')


def config_from_file(file):
    """Load configuration from file"""
    with open(file, encoding='utf-8') as file:
        return yaml.full_load(file)


def credentials_by_name(name, group):
    """Load credentials by evaluation name"""
    file = (
        reseval.EVALUATION_DIRECTORY /
        name /
        'credentials' /
        f'{group}.json')
    return credentials_from_file(file)


def credentials_from_file(file):
    """Load credentials corresponding to an evaluation from file"""
    with open(file) as file:
        return json.load(file)


def environment_variables_by_name(name):
    """Load environment variables corresponding to an evaluation"""
    file = reseval.EVALUATION_DIRECTORY / name / 'credentials' / '.env'

    # Raise if the file does not exist
    if not file.exists():
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), file)

    # Load as environment variables
    dotenv.load_dotenv(file, override=True)


###############################################################################
# Database table loading
###############################################################################


def conditions(name):
    """Load the table of conditions used in a subjective evaluation"""
    return load_table(name, 'conditions')


def participants(name):
    """Load the table of participants from a subjective evaluation"""
    return load_table(name, 'participants')


def responses(config):
    """Load the table of responses from a subjective evaluation"""
    return load_table(config, 'responses')


###############################################################################
# Utilities
###############################################################################


def load_table(name, table):
    """Load a database table that has been downloaded to local storage"""
    file = (
        reseval.EVALUATION_DIRECTORY /
        name /
        'tables' /
        f'{table}.csv')
    with open(file) as file:
        return [row for row in csv.DictReader(file)]



import argparse
from pathlib import Path

import reseval


###############################################################################
# Get the results of a subjective evaluation
###############################################################################


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(
        description='Get the results of a subjective evaluation')
    parser.add_argument(
        'name',
        help='The name of the subjective evaluation to retrieve results for')
    parser.add_argument(
        '--directory',
        type=Path,
        default=Path(),
        help='The directory to save results')
    return parser.parse_args()


reseval.results(**vars(parse_args()))



import argparse
from pathlib import Path

import reseval


###############################################################################
# Entry point for extending a subjective evaluation
###############################################################################


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(
        description='Extend a subjective evaluation')
    parser.add_argument('name', help='The name of the evaluation to extend')
    parser.add_argument(
        'participants',
        type=int,
        help='The number of new participants to add to the evaluation')
    parser.add_argument(
        '--local',
        action='store_true',
        help='Run subjective evaluation locally')
    parser.add_argument(
        '--production',
        action='store_true',
        help='Deploy the subjective evaluation to crowdsource participants')
    return parser.parse_args()


reseval.extend(**vars(parse_args()))



import json
import os
import shutil
import tempfile
import time
from pathlib import Path

import boto3

import reseval


###############################################################################
# AWS server management
###############################################################################


def create(name, detach=True):
    """Create an AWS server"""
    # Get unique identifier
    unique = reseval.load.credentials_by_name(name, 'unique')['unique']

    # Copy client and server to cache
    for path in reseval.ASSETS_DIR.rglob('*'):
        if path.is_dir():
            continue
        destination = reseval.CACHE / path.relative_to(reseval.ASSETS_DIR)
        destination.parent.mkdir(exist_ok=True, parents=True)
        shutil.copy(path, destination)

    # Install npm dependencies to build client
    client_directory = reseval.CACHE / 'client'
    client_directory.mkdir(exist_ok=True, parents=True)
    if not (client_directory / 'node_modules').exists():
        with reseval.chdir(client_directory):
            reseval.npm.install().wait()

    # Build client
    with reseval.chdir(reseval.CACHE / 'client'):
        reseval.npm.build().wait()

    # Connect to AWS
    client = connect()

    # Create server bundle
    with tempfile.TemporaryDirectory() as directory:
        directory = Path(directory)
        root = directory / 'reseval'
        root.mkdir()
        paths = [
            'client/build',
            'server',
            'package-lock.json',
            'package.json',
            'Procfile',
            'server.ts',
            'tsconfig.json']
        for path in paths:
            try:
                shutil.copytree(reseval.CACHE / path, root / path)
            except NotADirectoryError:
                shutil.copy(reseval.CACHE / path, root / path)

        # Zip
        shutil.make_archive(directory / 'reseval', 'zip', root)

        # Upload tarball
        reseval.storage.aws.upload(name, directory / 'reseval.zip')

        # Wait for upload to complete
        reseval.storage.aws.connect().get_waiter('object_exists').wait(
            Bucket=unique,
            Key='reseval.zip',
            WaiterConfig={'Delay': 2, 'MaxAttempts': 5})

        # Update application
        response = client.create_application_version(
            ApplicationName=unique,
            VersionLabel='Sample',
            SourceBundle={
                'S3Bucket': unique,
                'S3Key': 'reseval.zip'
            },
            AutoCreateApplication=True,
            Process=True)

        # Wait for build to finish
        response = client.describe_application_versions(
            ApplicationName=unique,
            VersionLabels=['Sample']
        )['ApplicationVersions'][0]
        while response['Status'].lower() in ['building', 'processing', 'unprocessed']:
            time.sleep(5)
            response = client.describe_application_versions(
                ApplicationName=unique,
                VersionLabels=['Sample']
            )['ApplicationVersions'][0]

    # Create elastic beanstalk environment
    client.create_environment(
        ApplicationName=unique,
        EnvironmentName=unique,
        SolutionStackName='64bit Amazon Linux 2 v5.8.3 running Node.js 18',
        OptionSettings=[{
            'Namespace': 'aws:autoscaling:launchconfiguration',
            'OptionName': 'IamInstanceProfile',
            'Value': 'aws-elasticbeanstalk-ec2-role'
        }])

    # Wait for environment build to finish
    response = client.describe_environments(
        ApplicationName=unique,
        EnvironmentNames=[unique])['Environments'][0]
    while response['Status'].lower() != 'ready':
        time.sleep(3)
        response = client.describe_environments(
            ApplicationName=unique,
            EnvironmentNames=[unique])['Environments'][0]

    # Process and deploy
    response = client.update_environment(
        ApplicationName=unique,
        EnvironmentName=unique,
        VersionLabel='Sample')

    # Wait for environment update to finish
    response = client.describe_environments(
        ApplicationName=unique,
        EnvironmentNames=[unique])['Environments'][0]
    while response['Status'].lower() != 'ready':
        time.sleep(3)
        response = client.describe_environments(
            ApplicationName=unique,
            EnvironmentNames=[unique])['Environments'][0]

    # Return application URL
    return {'URL': f'http://{response["CNAME"]}'}


def destroy(name, credentials):
    """Destroy an AWS server"""
    # Get unique identifier
    unique = reseval.load.credentials_by_name(name, 'unique')['unique']

    # Connect to AWS
    client = connect()

    # Get S3 bucket name
    arn = client.describe_application_versions(
            ApplicationName=unique,
            VersionLabels=['Sample']
        )['ApplicationVersions'][0]['ApplicationVersionArn']
    bucket = f'elasticbeanstalk-{arn.split(":")[3]}-{arn.split(":")[4]}'

    # Delete environment
    client.terminate_environment(
        EnvironmentName=unique,
        TerminateResources=True,
        ForceTerminate=True)

    # Delete application
    client.delete_application(ApplicationName=unique)

    # Connect to S3
    client = reseval.storage.aws.connect()

    try:

        # Get current bucket policy
        policy = json.loads(client.get_bucket_policy(Bucket=bucket)['Policy'])

        # Update bucket policy to allow deletion
        policy['Statement'] = [
            s for s in policy['Statement']
            if not (s['Action'] == 's3:DeleteBucket' and s['Effect'].lower() == 'deny')]
        client.put_bucket_policy(Bucket=bucket, Policy=json.dumps(policy))

        # Connect to S3 bucket
        bucket = boto3.Session(
            aws_access_key_id=os.environ['AWSAccessKeyId'],
            aws_secret_access_key=os.environ['AWSSecretKey']
        ).resource('s3').Bucket(bucket)

        # Delete contents
        bucket.objects.all().delete()

        # Delete bucket
        bucket.delete()

    except Exception:

        # Bucket was already deleted when environment was deleted
        pass


###############################################################################
# Utilities
###############################################################################


def configure(name, configuration):
    """Add configuration variables to server"""
    # Get unique identifier
    unique = reseval.load.credentials_by_name(name, 'unique')['unique']

    # Connect to AWS
    client = connect()

    # Update environment variables
    client.update_environment(
        ApplicationName=unique,
        EnvironmentName=unique,
        OptionSettings=[{
                'Namespace': 'aws:elasticbeanstalk:application:environment',
                'OptionName': key,
                'Value': value
            } for key, value in configuration.items()])

    # Wait for environment update to finish
    response = client.describe_environments(
        ApplicationName=unique,
        EnvironmentNames=[unique])['Environments'][0]
    while response['Status'].lower() != 'ready':
        time.sleep(3)
        response = client.describe_environments(
            ApplicationName=unique,
            EnvironmentNames=[unique])['Environments'][0]


def status(name):
    """Get current build status. One of ['succeeded', 'failed', 'pending']"""
    # Connect to AWS
    client = boto3.client('elasticbeanstalk')

    # Get environment info
    status = client.describe_environments(
        ApplicationName=name,
        EnvironmentNames=[name]
    )['Environments'][0]['Status']

    # Convert status descriptions to common format
    if status == 'Ready':
        return 'success'
    elif status == 'Launching':
        return 'pending'
    return 'failed'


###############################################################################
# Utilities
###############################################################################


def connect():
    """Connect to AWS"""
    # Load API keys into environment variables
    reseval.load.api_keys()

    # Add credentials and connect
    return boto3.Session(
        aws_access_key_id=os.environ['AWSAccessKeyId'],
        aws_secret_access_key=os.environ['AWSSecretKey']
    ).client('elasticbeanstalk', region_name='us-east-1')



import random
from string import ascii_lowercase, digits


def string(length, characters=ascii_lowercase + digits):
    """Generate a random string"""
    return ''.join(
        random.SystemRandom().choice(characters) for _ in range(length))



import json
import os

import boto3

import reseval


###############################################################################
# Constants
###############################################################################


# Cross-origin resource sharing
CORS = [
    {
        "AllowedHeaders": [
            "*"
        ],
        "AllowedMethods": [
            "GET"
        ],
        "AllowedOrigins": [
            "*"
        ],
        "ExposeHeaders": [],
        "MaxAgeSeconds": 3600
    }
]

# Public read policy
POLICY = {
    "Version": "2012-10-17",
    "Statement": [
        {
            "Sid": "PublicRead",
            "Effect": "Allow",
            "Principal": "*",
            "Action": [
                "s3:GetObject",
                "s3:GetObjectVersion"
            ],
            "Resource": "arn:aws:s3:::{}/*"
        }
    ]
}


###############################################################################
# AWS S3 file storage
###############################################################################


def create(config, directory):
    """Create an AWS S3 bucket for file storage"""
    name = config['name']

    # Get unique identifier
    unique = reseval.load.credentials_by_name(name, 'unique' )['unique']

    # Connect to S3
    client = connect()

    # Create bucket
    client.create_bucket(Bucket=unique, ObjectOwnership='ObjectWriter')

    # Set bucket policy to public read-only
    client.put_public_access_block(
        Bucket=unique,
        PublicAccessBlockConfiguration={
            'BlockPublicAcls': False,
            'IgnorePublicAcls': False,
            'BlockPublicPolicy': False,
            'RestrictPublicBuckets': False
        })
    client.put_bucket_acl(ACL='public-read', Bucket=unique)

    # Set bucket CORS to allow reads
    client.put_bucket_cors(
        Bucket=unique,
        CORSConfiguration={'CORSRules': CORS})

    # Set bucket policy to allow reads
    policy = POLICY.copy()
    policy['Statement'][0]['Resource'] = \
        policy['Statement'][0]['Resource'].format(unique)
    client.put_bucket_policy(Bucket=unique, Policy=json.dumps(policy))

    # Update client json
    with open(reseval.CLIENT_CONFIGURATION_FILE) as file:
        config = json.load(file)
    with open(reseval.CLIENT_CONFIGURATION_FILE, 'w') as file:
        json.dump(config | {'bucket': unique}, file, indent=4)

    # Upload evaluation files
    upload(name, directory)

    # Maybe upload listening test files
    if 'listening_test' in config:
        directory = reseval.LISTENING_TEST_DIRECTORY
        files = [item for item in directory.rglob('*') if not item.is_dir()]
        for file in files:
            destination = str('listening_test/' + file.name).replace('\\', '/')
            client.upload_file(
                str(file).replace('\\', '/'),
                unique,
                destination)


def destroy(name):
    """Delete an AWS S3 bucket"""
    # Get unique identifier
    unique = reseval.load.credentials_by_name(name, 'unique')['unique']

    # Load API keys as environment variables
    reseval.load.api_keys()

    # Connect to AWS
    bucket = boto3.Session(
        aws_access_key_id=os.environ['AWSAccessKeyId'],
        aws_secret_access_key=os.environ['AWSSecretKey']
    ).resource('s3').Bucket(unique)

    try:

        # Delete contents
        bucket.objects.all().delete()

        # Delete bucket
        bucket.delete()

    except Exception:

        # Bucket has already been deleted
        pass

    (
        reseval.EVALUATION_DIRECTORY /
        name /
        'credentials' /
        'storage.json'
    ).unlink(missing_ok=True)


def upload(name, file_or_directory):
    """Upload directory to AWS S3 bucket"""
    # Get bucket name
    unique = reseval.load.credentials_by_name(name, 'unique')['unique']

    # Connect to S3
    client = connect()

    # Upload directory
    if file_or_directory.is_dir():
        directory = file_or_directory
        files = [item for item in directory.rglob('*') if not item.is_dir()]
        for file in files:
            destination = str(file.relative_to(directory)).replace('\\', '/')
            client.upload_file(
                str(file).replace('\\', '/'),
                unique,
                destination)

    else:

        # Upload file
        file = file_or_directory
        destination = file.name
        client.upload_file(str(file), unique, file.name)

    # Return URL
    return f'https://{unique}.s3.amazonaws.com/{destination}'


###############################################################################
# Utilities
###############################################################################


def connect():
    """Connect to AWS"""
    # Load API keys into environment variables
    reseval.load.api_keys()

    # Add credentials and connect
    return boto3.Session(
        aws_access_key_id=os.environ['AWSAccessKeyId'],
        aws_secret_access_key=os.environ['AWSSecretKey']
    ).client('s3')



import datetime
import itertools
import json
import os
import re
from hashlib import md5
from pathlib import Path
import time

import boto3
import xmltodict as xmltodict
import uuid

import reseval


###############################################################################
# Constants
###############################################################################


# HIT preview URL
PREVIEW_URL = {
    'development': 'https://workersandbox.mturk.com/mturk/preview?groupId={}',
    'production': 'https://worker.mturk.com/mturk/preview?groupId={}'}

# HIT creation URL
URL = {
    'development': 'https://mturk-requester-sandbox.us-east-1.amazonaws.com',
    'production': 'https://mturk-requester.us-east-1.amazonaws.com'}

# Survey XML
XML_FILE = reseval.ASSETS_DIR / 'survey.xml'


###############################################################################
# Amazon Mechanical Turk crowdsourcing interface
###############################################################################


def active(config, credentials):
    """Returns True if the evaluation is still running"""
    return status(config, credentials) not in ['Reviewing', 'Reviewable']


def create(config, url, production=False):
    """Create a HIT"""
    # Connect to MTurk
    mturk = connect(production)

    # Get survey XML and link to our external survey
    with open(reseval.ASSETS_DIR / 'survey.xml') as file:
        question = file.read().format(url)

    # Create HIT
    cfg = config['crowdsource']
    hit = mturk.create_hit(
        Title=cfg['title'],
        Description=cfg['description'],
        Keywords=cfg['keywords'],
        Reward=str(cfg['payment']['base']),
        MaxAssignments=config['participants'],
        LifetimeInSeconds=cfg['duration']['total'],
        AssignmentDurationInSeconds=cfg['duration']['assignment'],
        AutoApprovalDelayInSeconds=cfg['duration']['autoapprove'],
        QualificationRequirements=qualifications(config),
        Question=question)

    # Log HIT
    hit_id = hit['HIT']['HITId']
    url_key = 'production' if production else 'development'
    preview_url = PREVIEW_URL[url_key].format(hit['HIT']['HITGroupId'])
    print(f'Created HIT {hit_id}. You can view your HIT at {preview_url}')

    # Return crowdsource credentials
    return {'HIT_ID': hit_id, 'PRODUCTION': production}


def destroy(config, credentials):
    """Delete a HIT"""
    if credentials['HIT_ID'] in list_hits(credentials):
        # Connect to MTurk
        mturk = connect(credentials['PRODUCTION'])

        # Stop HIT
        mturk.update_expiration_for_hit(
            HITId=credentials['HIT_ID'],
            ExpireAt=datetime.datetime(2000, 1, 1).timestamp())

        # The HIT can only be deleted once the status is one of 'Reviewing' or
        # 'Reviewable'. Otherwise, an error will be thrown. Sometimes, it takes
        # a couple of seconds after the update call above for the status to
        # change to 'Reviewable'. This loop is a time buffer that gives MTurk
        # 90 seconds to update the status. If the status doesn't change, the
        # most likely explanation is that someone is still taking the
        # evaluation.
        start = datetime.datetime.now()
        while (datetime.datetime.now() - start).total_seconds() < 90:
            if not active(config, credentials):
                break
            time.sleep(5)

        # Delete HIT
        mturk.delete_hit(HITId=credentials['HIT_ID'])


def extend(credentials, participants, name):
    """Extend a HIT with additional assignments"""
    # Connect to MTurk
    mturk = connect(credentials['PRODUCTION'])

    # Location to save unique response token
    token_file = Path(reseval.EVALUATION_DIRECTORY / name / 'tokens.json')

    try:

        # Last response was bad. Retry using same token.
        with open(token_file) as json_file:
            unique_token = json.load(json_file)['extend_token']

    except FileNotFoundError:

        # Create and save a new unique token
        unique_token = str(uuid.uuid4())
        data = {'extend_token': unique_token}
        with open(token_file, 'w') as json_file:
            json.dump(data, json_file)

    try:

        # Extend HIT
        mturk.create_additional_assignments_for_hit(
            HITId=credentials['HIT_ID'],
            UniqueRequestToken=unique_token,
            NumberOfAdditionalAssignments=participants)

    except Exception as exception:

        # If the token is in the exception args, we assume the token has been
        # used and delete the token
        if re.search(re.escape(unique_token), exception.args[0]) is not None:
            token_file.unlink(True)

        else:
            raise exception


def exists(config, credentials):
    """Returns true if the evaluation exists"""
    return credentials['HIT_ID'] in list_hits(credentials)


def paid(credentials):
    """Returns True if all participants have been paid"""
    return not assignments(credentials, ['Submitted'])


def pay(config, credentials):
    """Evaluate and pay participants"""
    # Get assignment IDs and completion codes
    mturk_results = results(credentials)

    # Get participants and responses from database
    reseval.database.download(
        config['name'],
        reseval.EVALUATION_DIRECTORY / config['name'] / 'tables',
        ['participants', 'responses'])
    participants = reseval.load.participants(config['name'])
    responses = reseval.load.responses(config['name'])

    # Match participants with responses
    participant_responses = {}
    for participant in participants:
        pid = participant['ID']
        participant_responses[pid] = (
                participant |
                {'responses': list(filter(
                    lambda x: x['Participant'] == pid,
                    responses))})

    # Iterate over participants
    for result in mturk_results:

        # Find matching completion code
        match = None
        for pid in participant_responses:
            participant_code = participant_responses[pid]['CompletionCode']
            completion_code = result['completion_code']
            if participant_code == completion_code:
                match = pid

        # Only process payment if the participant has not already been paid
        if result['status'] == 'Submitted':

            # Reject work if the completion codes do not match
            if match is None and result['status'] == 'Submitted':
                reject(
                    credentials,
                    result['assignment_id'],
                    'Survey completion code does not match')
                block(
                    config,
                    credentials,
                    result['worker_id'],
                    'Survey completion code does not match')
                continue

            # Approve work
            approve(credentials, result['assignment_id'])

            # Block workers who do not complete evaluations
            if (
                len(participant_responses[match]['responses']) !=
                config['samples_per_participant']
            ):
                block(
                    config,
                    credentials,
                    result['worker_id'],
                    'Incomplete evaluation')

            # If they passed prescreening and completed evaluation, give
            # the participant a bonus
            else:
                bonus(config, credentials, result['assignment_id'])


def progress(credentials):
    """Retrieve the number of participants that have taken the evaluation"""
    return len(assignments(credentials))


###############################################################################
# Utilities
###############################################################################


def approve(credentials, assignment_id):
    """Approve an assignment"""
    # Connect to MTurk
    mturk = connect(credentials['PRODUCTION'])

    # Skip if already processed
    assignment = mturk.get_assignment(AssignmentId=assignment_id)
    if assignment['Assignment']['AssignmentStatus'] == 'Submitted':
        # Approve assignment
        mturk.approve_assignment(AssignmentId=assignment_id)


def assignments(credentials, statuses=None):
    """Retrieve a list of all assignments for a HIT"""
    # Default to all statuses
    statuses = (
        ['Submitted', 'Approved', 'Rejected']
        if statuses is None else statuses)

    # Connect to MTurk
    mturk = connect(credentials['PRODUCTION'])

    # Download results
    paginator = mturk.get_paginator('list_assignments_for_hit')
    iterator = paginator.paginate(
        HITId=credentials['HIT_ID'],
        AssignmentStatuses=statuses)

    # Get assignments
    return list(itertools.chain.from_iterable(
        page['Assignments'] for page in iterator))


def block(config, credentials, worker_id, reason):
    """Block a worker"""
    # Connect to MTurk
    mturk = connect(credentials['PRODUCTION'])

    # Block worker
    mturk.create_worker_block(WorkerId=worker_id, Reason=reason)


def bonus(config, credentials, assignment_id):
    """Give a participant a bonus"""
    # We only send one bonus per assignment, so we use the assignment ID to
    # generate the unique token
    unique_token = md5(assignment_id.encode('utf-8')).hexdigest()

    # Connect to MTurk
    mturk = connect(credentials['PRODUCTION'])

    # Get worker ID
    assignment = mturk.get_assignment(AssignmentId=assignment_id)
    worker_id = assignment['Assignment']['WorkerId']

    try:

        # Approve assignment
        mturk.send_bonus(
            WorkerId=worker_id,
            BonusAmount=str(config['crowdsource']['payment']['completion']),
            AssignmentId=assignment_id,
            UniqueRequestToken=unique_token,
            Reason='Passed prescreening and completed evaluation. Thank you!')

    except Exception as exception:

        # If the token is in the exception args, we assume the worker has
        # already been given a bonus
        if re.search(re.escape(unique_token), exception.args[0]) is None:
            raise exception


def connect(production=False):
    """Connect to MTurk"""
    # Load API keys
    reseval.load.api_keys()

    # Connect to MTurk
    return boto3.Session(
        aws_access_key_id=os.environ['AWSAccessKeyId'],
        aws_secret_access_key=os.environ['AWSSecretKey']
    ).client(
        'mturk',
        region_name='us-east-1',
        endpoint_url=URL['production' if production else 'development'])


def list_hits(credentials):
    """List the IDs of all HITs on MTurk"""
    # Connect to S3
    mturk = connect(credentials['PRODUCTION'])

    # Get list of HITs
    iterator = mturk.get_paginator('list_hits').paginate()
    responses = list(iterator)

    # Extract HIT Ids
    if responses:
        return list(itertools.chain.from_iterable([
            [hit['HITId'] for hit in response['HITs']]
            for response in responses]))

    return []


def qualifications(config):
    """Format participant qualificiations"""
    qualifications = []

    # Country of origin
    cfg = config['crowdsource']['filter']
    if 'countries' in cfg:
        locales = [{'Country': country} for country in cfg['countries']]
        qualifications.append({
            'QualificationTypeId': '00000000000000000071',
            'Comparator': 'In',
            'LocaleValues': locales,
            'RequiredToPreview': True})

    # Number of approved tasks
    if 'approved_tasks' in cfg and cfg['approved_tasks'] > 0:
        qualifications.append({
            'QualificationTypeId': '00000000000000000040',
            'Comparator': 'GreaterThan',
            'IntegerValues': [cfg['approved_tasks']],
            'RequiredToPreview': True})

    # Approval rating
    if 'approval_rating' in cfg and cfg['approval_rating'] > 0:
        qualifications.append({
            'QualificationTypeId': '000000000000000000L0',
            'Comparator': 'GreaterThanOrEqualTo',
            'IntegerValues': [cfg['approval_rating']],
            'RequiredToPreview': True})

    return qualifications


def reject(credentials, assignment_id, reason):
    """Reject assignment by ID"""
    # Connect to MTurk
    mturk = connect(credentials['PRODUCTION'])

    # Skip if already processed
    assignment = mturk.get_assignment(AssignmentId=assignment_id)
    if assignment['Assignment']['AssignmentStatus'] == 'Submitted':
        # Reject assignment
        mturk.reject_assignment(
            AssignmentId=assignment_id,
            RequesterFeedback=reason)


def results(credentials):
    """Get a list of all assignment IDs and completion codes"""
    result = []
    for assignment in assignments(credentials):
        # Parse XML
        xml_doc = xmltodict.parse(assignment['Answer'])

        # Get completion code
        completion_code = xml_doc['QuestionFormAnswers']['Answer']['FreeText']

        # Format result
        result.append({
            'assignment_id': assignment['AssignmentId'],
            'completion_code': completion_code,
            'status': assignment['AssignmentStatus'],
            'worker_id': assignment['WorkerId']})

    return result


def status(config, credentials):
    """Get the status of the current HIT"""
    # Connect to MTurk
    mturk = connect(credentials['PRODUCTION'])

    # Get HIT status
    return mturk.get_hit(HITId=credentials['HIT_ID'])['HIT']['HITStatus']



import dotenv

import reseval


###############################################################################
# Credential management
###############################################################################


def credentials(
    aws_api_key=None,
    aws_api_secret_key=None,
    heroku_api_key=None,
    mysql_local_user=None,
    mysql_local_password=None):
    """Add or update credentials"""
    # Update credentials for local development
    local_credentials(mysql_local_user, mysql_local_password)

    # Update API keys for remote development
    remote_credentials(aws_api_key, aws_api_secret_key, heroku_api_key)


###############################################################################
# Utilities
###############################################################################


def local_credentials(mysql_local_user=None, mysql_local_password=None):
    """Update database credentials for local development"""
    try:

        # Load existing credentials
        creds = dotenv.dotenv_values(reseval.ENVIRONMENT_FILE)

    except FileNotFoundError:

        # No credentials found
        creds = {}

    # Local development assumes localhost
    creds['MYSQL_HOST'] = 'localhost'

    # Update credentials
    if mysql_local_user is not None:
        creds['MYSQL_USER'] = mysql_local_user
    if mysql_local_password is not None:
        creds['MYSQL_PASS'] = mysql_local_password

    # Write to environment file
    reseval.ENVIRONMENT_FILE.parent.mkdir(exist_ok=True, parents=True)
    with open(reseval.ENVIRONMENT_FILE, 'w') as file:
        for key, value in creds.items():
            file.write(f'{key}={value}\n')


def remote_credentials(
    aws_api_key=None,
    aws_api_secret_key=None,
    heroku_api_key=None):
    """Update API keys for remote development"""
    if reseval.KEYS_FILE.exists():

        # Load existing credentials
        with open(reseval.KEYS_FILE) as file:
            lines = [line.rstrip() for line in file]
        keys = dict([line.split('=') for line in lines])

    else:

        # No credentials found
        keys = {}

    # Add credentials
    if aws_api_key is not None:
        keys['AWSAccessKeyId'] = aws_api_key
    if aws_api_secret_key is not None:
        keys['AWSSecretKey'] = aws_api_secret_key
    if heroku_api_key is not None:
        keys['HerokuAccessKey'] = heroku_api_key

    # Save credentials
    with open(reseval.KEYS_FILE, 'w') as file:
        for key, value in keys.items():
            file.write(f'{key}={value}\n')



class ValidationError(Exception):

    def __init__():
        super().__init__(
            'Validation check failed. Please check that the directory ' +
            'structure of the evaluation files is correct for the type ' +
            'of test you are running.')



import argparse
from pathlib import Path

import reseval


###############################################################################
# File conversion
###############################################################################


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(
        description='Perform file compression recursively on a directory')
    parser.add_argument(
        'input_directory',
        type=Path,
        help='The directory of files to compress')
    parser.add_argument(
        'output_directory',
        type=Path,
        help='The location to save compressed files')


reseval.convert(**vars(parse_args()))



import argparse
import glob
import os
import shutil
import subprocess
from pathlib import Path


###############################################################################
# Convert audio to mp3
###############################################################################


def from_file_to_file(input_file, output_file=None, verbose=True):
    """Convert audio file to mp3"""
    # Handle input files starting with hyphen
    clean_input = False
    if input_file.stem.startswith('-'):
        dummy_file = input_file.parent / input_file.name[1:]
        shutil.copyfile(input_file, dummy_file)
        input_file = dummy_file
        clean_input = True

    # Handle output files starting with hyphen
    clean_output = False
    if output_file.stem.startswith('-'):
        output_file = output_file.parent / output_file.name[1:]
        clean_output = True

    # Default output filename is same as input but with MP3 extension
    if output_file is None:
        output_file = input_file.with_suffix('.mp3')

    # Ensure output directory exists
    output_file.parent.mkdir(exist_ok=True, parents=True)

    # Convert
    args = [
        'ffmpeg',
        '-y',
        '-i',
        str(input_file),
        '-b:a',
         '320k',
        str(output_file)]
    process = subprocess.Popen(
        args,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True)
    stdout, stderr = process.communicate()

    # Maybe print
    if verbose or process.returncode != 0:
        print(stdout)
        print(stderr)

    # Clean-up input files starting with hyphen
    if clean_input:
        os.remove(input_file)

    # Clean-up output files starting with hyphen
    if clean_output:
        os.replace(output_file, output_file.parent / ('-' + output_file.name))


def from_files_to_files(input_files, output_files=None):
    """Convert audio files to mp3"""
    # Convert to paths
    input_files = [Path(file) for file in input_files]

    # Default output filename is same as input but with MP3 extension
    if output_files is None:
        output_files = [file.with_suffix('.mp3') for file in input_files]

    # Multiprocess conversion
    # with mp.Pool() as pool:
    #     pool.starmap(from_file_to_file, zip(input_files, output_files))

    for input_file, output_file in zip(input_files, output_files):
        from_file_to_file(input_file, output_file)


###############################################################################
# Entry point
###############################################################################


def expand_files(files):
    """Expands a wildcard to a list of paths for Windows compatibility"""
    # Split at whitespace
    files = files.split()

    # Handle wildcard expansion
    if len(files) == 1 and '*' in files[0]:
        files = glob.glob(files[0])

    # Convert to Path objects
    return files


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(description='Convert audio to MP3')

    # Handle wildcards across platforms
    if os.name == 'nt':
        parser.add_argument(
            '--input_files',
            type=expand_files,
            help='The audio files to convert to mp3')
    else:
        parser.add_argument(
            '--input_files',
            nargs='+',
            help='The audio files to convert to mp3')

    parser.add_argument(
        '--output_files',
        type=Path,
        nargs='+',
        help='The corresponding output files. ' + \
             'Uses same filename with mp3 extension by default')
    return parser.parse_args()


if __name__ == '__main__':
    from_files_to_files(**vars(parse_args()))



import re

import reseval


###############################################################################
# Heroku ClearDB database management
###############################################################################


def create(name):
    """Create a MySQL database on Heroku"""
    # Get the heroku application
    app = reseval.app.heroku.get(name)

    # Add the ClearDB MySQL database add-on
    app.install_addon(plan_id_or_name='cleardb:ignite', config={})

    # Retrieve the ClearDB URL
    url = app.config()['CLEARDB_DATABASE_URL']

    # Parse URL to obtain credentials
    user, password, host, unique = re.split('[/@?:]', url[8:])[:4]
    credentials = {
        'MYSQL_DBNAME': unique,
        'MYSQL_HOST': host,
        'MYSQL_USER': user,
        'MYSQL_PASS': password}

    return credentials


def destroy(name):
    """Destroy a MySQL database on Heroku"""
    reseval.app.heroku.destroy(name)



import os

import heroku3

import reseval


def configure(name, key, value):
    """Add an environment variable to a Heroku application"""
    # Get Heroku application
    app = get(name)

    # Set environment variable
    app.config()[key] = value


def connect():
    """Connect to Heroku"""
    reseval.load.api_keys()
    return heroku3.from_key(os.environ['HerokuAccessKey'])


def create(name):
    """Create a Heroku web application"""
    # Get unique identifier
    unique = reseval.load.credentials_by_name(name, 'unique')['unique']

    # Create Heroku app
    connect().create_app(name=unique)


def destroy(name):
    """Destroy a Heroku app"""
    # Get unique identifier
    unique = reseval.load.credentials_by_name(name, 'unique')['unique']

    try:

        # Destroy app
        get(unique).delete()

    except (FileNotFoundError, KeyError):

        # Handle app not existing
        pass

    # Delete credentials file
    (
        reseval.EVALUATION_DIRECTORY /
        name /
        'credentials' /
        'app.json'
    ).unlink(missing_ok=True)


def get(name):
    """Retrieve a Heroku application"""
    # Load application credentials
    unique = reseval.load.credentials_by_name(name, 'unique')['unique']

    # Get Heroku application
    return connect().apps()[unique]



import json
import os
import shutil
import string
import typing

import reseval


###############################################################################
# Create a subjective evaluation
###############################################################################


def create(
    config: typing.Union[str, bytes, os.PathLike],
    directory: typing.Union[str, bytes, os.PathLike],
    local: bool = False,
    production: bool = False,
    detach: bool = False) -> str:
    """Setup a subjective evaluation

    Args:
        config: The configuration file
        directory: The directory containing the files to evaluate
        local: Run subjective evaluation locally
        production: Deploy the subjective evaluation to crowdsource
            participants
        detach: If running locally, detaches the server process

    Returns:
        str: The name of the evaluation, as given in the configuration file
    """
    if local and production:
        raise ValueError('Cannot deploy production build locally')

    # Load configuration file
    cfg = reseval.load.config_from_file(config)
    name = cfg['name']

    if (reseval.EVALUATION_DIRECTORY / name).exists():
        raise ValueError(f'Evaluation {name} already exists')

    # Copy client and server to cache
    for path in reseval.ASSETS_DIR.rglob('*'):
        if path.is_dir():
            continue
        destination = reseval.CACHE / path.relative_to(reseval.ASSETS_DIR)
        destination.parent.mkdir(exist_ok=True, parents=True)
        shutil.copy(path, destination)

    # Don't overwrite production data
    if not production:
        prod_file = reseval.EVALUATION_DIRECTORY / name / '.prod'
        if prod_file.exists():
            raise ValueError(
                f'Not overwriting results of evaluation {name}',
                 'which has been run in production')

    # Save configuration as json for the frontend to access
    with open(reseval.CLIENT_CONFIGURATION_FILE, 'w') as file:
        json.dump(cfg | {'local': local}, file, indent=4)

    # Create test
    test = reseval.test.get(cfg)(cfg, directory)

    # Create and save assignments
    with open(reseval.CLIENT_ASSIGNMENT_FILE, 'w') as file:
        json.dump(test.assign(cfg['random_seed']), file, indent=4)

    # MOS test requires deterministic conditions
    if cfg['test'] == 'mos':
        conditions = test.assign_conditions(cfg['random_seed'])
    else:
        conditions = []
    with open(reseval.CLIENT_CONDITION_FILE, 'w') as file:
        json.dump(conditions, file, indent=4)

    # Copy configuration file
    file = reseval.EVALUATION_DIRECTORY / name / 'config.yaml'
    file.parent.mkdir(exist_ok=True, parents=True)
    try:
        shutil.copyfile(config, file)
    except shutil.SameFileError:
        pass

    # Indicate if we're running local development
    local_file = reseval.EVALUATION_DIRECTORY / name / '.local'
    if local:
        local_file.touch()
    elif local_file.exists():
        local_file.unlink()

    try:

        # Create a globally unique name to prevent collision
        unique = (
            reseval.random.string(1, string.ascii_lowercase) +
            reseval.random.string(23))

        # Save unique name
        credentials_file = (
            reseval.EVALUATION_DIRECTORY /
            name /
            'credentials' /
            'unique.json')
        credentials_file.parent.mkdir(exist_ok=True, parents=True)
        with open(credentials_file, 'w') as file:
            json.dump({'unique': unique}, file)

        # Create file storage and upload
        reseval.storage.create(cfg, directory, local)

        if local:

            # Create database before server
            credentials = reseval.database.create(cfg, test, local)
            url = reseval.server.create(cfg, local, detach=detach)

        else:

            # If heroku is used for either the database or server, setup the app here
            if cfg['server'] == 'heroku' or cfg['database'] == 'heroku':
                reseval.app.heroku.create(name)

            # Create server
            url = reseval.server.create(cfg, local, detach=detach)

            # Create database
            credentials = reseval.database.create(cfg, test, local)

            # Add database environment variables to server
            if cfg['server'] == 'aws':
                reseval.server.aws.configure(name, credentials)
            elif cfg['server'] == 'heroku':
                for key, value in credentials.items():
                    reseval.app.heroku.configure(name, key, value)

        # Launch crowdsourced evaluation
        reseval.crowdsource.create(cfg, url, local, production)

    except (Exception, KeyboardInterrupt) as exception:

        # Cleanup resources
        reseval.destroy(name, True)

        raise exception

    # Indicate that the evaluation has been run in production
    if production:
        prod_file = reseval.EVALUATION_DIRECTORY / name / '.prod'
        with open(prod_file, 'w') as file:
            pass

    # Return the evaluation name for, e.g., monitoring and analysis
    return name



import json

import reseval


###############################################################################
# Crowdsourcing interface
###############################################################################


def active(name):
    """Returns True if the evaluation is still running"""
    # Get config
    config = reseval.load.config_by_name(name)

    # Local dev is active until we have enough participants and responses
    if reseval.is_local(name):
        try:
            reseval.database.download(
                name,
                reseval.EVALUATION_DIRECTORY / name / 'tables',
                ['responses'])
            responses = reseval.load.responses(name)
            participants = len(set(
                response['Participant'] for response in responses))
            samples = \
                config['participants'] * config['samples_per_participant']
            return (
                len(responses) < samples or
                participants < config['participants'])
        except FileNotFoundError:
            return True

    # Get credentials
    try:
        credentials = reseval.load.credentials_by_name(name, 'crowdsource')
    except FileNotFoundError:
        return False

    # Check if the evaluation is active
    return module(config).active(config, credentials)


def assignments(name):
    """Get the assignment results"""
    # Skip if performing local development
    if reseval.is_local(name):
        return []

    # Get config
    config = reseval.load.config_by_name(name)

    # Get credentials
    credentials = reseval.load.credentials_by_name(name, 'crowdsource')

    # Extend task
    return module(config).assignments(credentials)


def create(config, url, local=False, production=False):
    """Create a subjective evaluation"""
    # Skip if performing local development
    if local:
        return

    # Skip if evaluation exists
    if exists(config['name']):
        raise ValueError('Not overwriting existing evaluation')

    print('Creating crowdsource task...')

    # Create crowdsource task
    credentials = module(config).create(config, url, production)

    # Save crowdsource credentials
    file = (
        reseval.EVALUATION_DIRECTORY /
        config['name'] /
        'credentials' /
        'crowdsource.json')
    file.parent.mkdir(exist_ok=True, parents=True)
    with open(file, 'w') as file:
        json.dump(credentials, file, indent=4, sort_keys=True)


def destroy(name):
    """Delete a subjective evaluation"""
    try:

        # Skip if performing local development
        if reseval.is_local(name):
            return

        # Get config
        config = reseval.load.config_by_name(name)

        # Get credentials
        credentials = reseval.load.credentials_by_name(name, 'crowdsource')

        # Destroy crowdsource task
        module(config).destroy(config, credentials)

    except FileNotFoundError:

        pass

    # Clean-up credentials
    (
        reseval.EVALUATION_DIRECTORY /
        name /
        'credentials' /
        'crowdsource.json'
    ).unlink(missing_ok=True)


def extend(name, participants):
    """Extend a subjectve evaluation"""
    # Skip if performing local development
    if reseval.is_local(name):
        return

    # Get config
    config = reseval.load.config_by_name(name)

    # Get credentials
    credentials = reseval.load.credentials_by_name(name, 'crowdsource')

    # Extend task
    module(config).extend(credentials, participants, name)


def exists(name):
    """Returns true if the evaluation exists"""
    if reseval.is_local(name):
        return False

    try:

        # Get config
        config = reseval.load.config_by_name(name)

        # Get credentials
        credentials = reseval.load.credentials_by_name(name, 'crowdsource')

    except FileNotFoundError:

        return False

    # Check if evaluation exists
    return module(config).exists(config, credentials)


def paid(name):
    """Check if all participants have been paid"""
    # Skip if performing local development
    if reseval.is_local(name):
        return True

    # Get config
    config = reseval.load.config_by_name(name)

    try:

        # Get credentials
        credentials = reseval.load.credentials_by_name(name, 'crowdsource')

    except FileNotFoundError:

        # We assume that the evaluation was never started or already finished
        # if we do not have the credentials
        return True

    # Check if participants have been paid
    return module(config).paid(credentials)


def pay(name):
    """Pay participants"""
    # Skip if performing local development
    if reseval.is_local(name):
        return

    # Get config
    config = reseval.load.config_by_name(name)

    # Get credentials
    credentials = reseval.load.credentials_by_name(name, 'crowdsource')

    # Pay participants
    module(config).pay(config, credentials)


def progress(name):
    """Retrieve the number of participants who have taken the evaluation"""
    # Skip if performing local development
    if reseval.is_local(name):
        return

    # Get config
    config = reseval.load.config_by_name(name)

    # Get credentials
    credentials = reseval.load.credentials_by_name(name, 'crowdsource')

    # Check current progress
    return module(config).progress(credentials)


###############################################################################
# Utilities
###############################################################################


def module(config):
    """Get the crowdsourcing module to use"""
    platform = config['crowdsource']['platform']
    if platform == 'mturk':
        return reseval.crowdsource.mturk
    raise ValueError(f'Crowdsource platform {platform} not recognized')



from pathlib import Path
from setuptools import find_packages, setup


def long_description():
    """Generate description to post on PyPi"""
    with open('README.md', encoding='utf8') as file:
        return file.read()


def package_data():
    """Get the non-python files that should be included in release"""
    module_directory = Path(__file__).parent / 'reseval'
    assets_directory = module_directory / 'assets'

    # Package data should be paths relative to package
    return {'reseval': [
        str(path.relative_to(module_directory))
        for path in assets_directory.rglob('*')]}


# Setup
setup(
    name='reseval',
    description='Reproducible Subjective Evaluation',
    version='0.1.6',
    author='Max Morrison, Brian Tang, Gefei Tan, Bryan Pardo',
    author_email='maxrmorrison@gmail.com',
    url='https://github.com/reseval/reseval',
    install_requires=[
        'appdirs>=1.4.4',
        'boto3>=1.21.3',
        'matplotlib>=3.5.2',
        'mysql-connector-python>=8.0.28',
        'psutil>=5.9.0',
        'python-dotenv>=0.19.2',
        'pyyaml>=6.0',
        'rich>=11.2.0',
        'scipy>=1.8.0',
        'xmltodict>=0.12.0',
        'heroku3>=5.1.4'],
    packages=find_packages(),
    package_data=package_data(),
    long_description=long_description(),
    long_description_content_type='text/markdown',
    keywords=[
        'annotation',
        'audio',
        'ab',
        'abx',
        'crowdsourcing',
        'evaluation',
        'image',
        'mos',
        'mushra',
        'speech',
        'subjective'])



from .constants import *
from .core import *
from .error import *
from .create import create
from .credentials import credentials
from .extend import extend
from .destroy import destroy
from .monitor import monitor
from .results import results
from .pay import pay
from . import app
from . import convert
from . import crowdsource
from . import database
from . import filter
from . import load
from . import npm
from . import plot
from . import random
from . import server
from . import stats
from . import storage
from . import test



import shutil

import reseval


###############################################################################
# Client file storage management
###############################################################################


def create(config, directory):
    """Add evaluation files to client storage"""
    for path in directory.iterdir():
        upload(config['name'], path)

    # Add listening test files to client storage
    if 'listening_test' in config:
        upload(config['name'], reseval.LISTENING_TEST_DIRECTORY)


def destroy(name):
    """Remove all evaluation files in client public directory"""
    try:
        for path in reseval.CLIENT_PUBLIC_DIRECTORY.iterdir():
            if path.name != '.gitkeep':
                try:
                    shutil.rmtree(path)
                except NotADirectoryError:
                    path.unlink()
    except FileNotFoundError:
        pass


def upload(name, file_or_directory):
    """Upload a file or directory to client storage"""
    destination = reseval.CLIENT_PUBLIC_DIRECTORY / file_or_directory.name
    reseval.CLIENT_PUBLIC_DIRECTORY.mkdir(exist_ok=True, parents=True)

    # Upload directory
    if file_or_directory.is_dir():
        shutil.copytree(file_or_directory, destination)

    # Upload file
    else:
        shutil.copyfile(file_or_directory, destination)

    # Return URL
    return f'/evaluation-files/{file_or_directory.name}'



from pathlib import Path

import appdirs


###############################################################################
# Cache
###############################################################################


# Application data directory. This is a persistant cross-platform cache
# directory. We use this to store a copy of the client and server code, as
# well as the following locally generated files.
# - .env - The local database credentials used for local development
# - .keys - The API keys
# - client/build - The optimized front-end code for deployment
# - client/node_modules - The installed front-end packages
# - client/src/json/assignments.json - The pre-generated assignments of
#     evaluators to files to evaluate
# - client/src/json/config.json - The configuration file including any
#     additions made by reseval.create, such as whether we are running local
#     development
# - evaluations - A directory containing the credentials, database output, and
#     results of every evaluation. Credentials are deleted when the
#     corresponding resource is destroyed.
# - node_modules - The installed server packages used for local development
CACHE = Path(appdirs.user_cache_dir('reseval', 'reseval'))


###############################################################################
# Credentials
###############################################################################


# Environment variable file
ENVIRONMENT_FILE = CACHE / '.env'

# API keys file
KEYS_FILE = CACHE / '.keys'


###############################################################################
# Directories
###############################################################################


# Location of Python package data
ASSETS_DIR = Path(__file__).parent / 'assets'

# Location that the client will load JSON files that we save
CLIENT_JSON_DIRECTORY = CACHE / 'client' / 'src' / 'json'

# Location that the client expects to find an assignment JSON file
CLIENT_ASSIGNMENT_FILE = CLIENT_JSON_DIRECTORY / 'assignments.json'

# Location that the client expects to find a JSON file of condition assignments
CLIENT_CONDITION_FILE = CLIENT_JSON_DIRECTORY / 'conditions.json'

# Location that the client expects to find a configuration JSON file
CLIENT_CONFIGURATION_FILE = CLIENT_JSON_DIRECTORY / 'config.json'

# Location that the client serves files from if we are not using cloud storage
CLIENT_PUBLIC_DIRECTORY = (
    CACHE /
    'client' /
    'public' /
    'evaluation-files')

# Location to save results
EVALUATION_DIRECTORY = CACHE / 'evaluations'

# Location of listening test audio files
LISTENING_TEST_DIRECTORY = ASSETS_DIR / 'listening_test'


###############################################################################
# Server
###############################################################################


# URL used for local deployment
LOCALHOST_URL = 'http://localhost:3001/'



import json
import psutil
import webbrowser

import requests
from requests.adapters import HTTPAdapter, Retry

import reseval


###############################################################################
# Local server management
###############################################################################


def create(name, detach=False):
    """Deploy a local server"""
    # Maybe install server
    if not (reseval.CACHE / 'node_modules').exists():
        with reseval.chdir(reseval.CACHE):
            reseval.npm.install().wait()

    # Maybe install client
    client_directory = reseval.CACHE / 'client'
    if not (client_directory / 'node_modules').exists():
        with reseval.chdir(client_directory):
            reseval.npm.install().wait()

    # Launch server process
    with reseval.chdir(reseval.CACHE):
        process = reseval.npm.start()

    # Ping server until we get a response
    session = requests.Session()
    retries = Retry(total=11,
                    backoff_factor=0.1,
                    status_forcelist=[500, 502, 503, 504])
    session.mount('http://', HTTPAdapter(max_retries=retries))
    session.get(reseval.LOCALHOST_URL)

    # Open web browser to evaluation
    webbrowser.open(reseval.LOCALHOST_URL)

    # Save server credentials
    credentials = {'URL': reseval.LOCALHOST_URL, 'PID': process.pid}
    file = (
        reseval.EVALUATION_DIRECTORY /
        name /
        'credentials' /
        'server.json')
    file.parent.mkdir(exist_ok=True, parents=True)
    with open(file, 'w') as file:
        json.dump(credentials, file, indent=4, sort_keys=True)

    # Maybe wait for process to finish
    if not detach:
        process.wait()

    return credentials


def destroy(name, credentials):
    """Shutdown a local server"""
    try:
        parent = psutil.Process(credentials['PID'])
        for child in parent.children(recursive=True):
            child.kill()
        parent.kill()
    except psutil.NoSuchProcess:
        pass



import argparse

import reseval


###############################################################################
# Entry point for adding or updating credentials
###############################################################################


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(
        description='Stop a subjective evaluation without ' +
                    'destroying resources')
    parser.add_argument('--aws_api_key', help='The public API key for AWS')
    parser.add_argument(
        '--aws_api_secret_key',
        help='The private API key for AWS')
    parser.add_argument('--heroku_api_key', help='The API key for Heroku')
    parser.add_argument(
        '--mysql_local_user',
        help='The username of the local MySQL database')
    parser.add_argument(
        '--mysql_local_password',
        help='The corresponding password of the local MySQL database')
    return parser.parse_args()


reseval.credentials(**vars(parse_args()))



import reseval


###############################################################################
# Test selection
###############################################################################


def get(config):
    """Get a test class"""
    test = config['test']

    if test == 'ab':
        return reseval.test.AB
    elif test == 'abx':
        return reseval.test.ABX
    elif test == 'mos':
        return reseval.test.MOS
    elif test == 'mushra':
        return reseval.test.MUSHRA
    elif test == 'wordselect':
        return reseval.test.WordSelect
    raise ValueError(f'Test {test} is not defined')



import argparse
import glob
import io
import multiprocessing as mp
import os
from pathlib import Path

from PIL import Image


###############################################################################
# Convert image to jpeg
###############################################################################


def from_file_to_file(input_file, output_file=None):
    """Convert image file to jpeg"""
    # Default output filename is same as input but with JPEG extension
    if output_file is None:
        output_file = input_file.with_suffix('.jpg')

    # Open image file
    image = Image.open(input_file)

    # Create raw byte buffer
    buffer = io.BytesIO()

    # Perform compression to 25% of the original file size
    image.save(buffer, 'JPEG', quality=25)

    # Write the buffer to a file
    with open(output_file, 'w') as file:
        file.write(buffer.contents())


def from_files_to_files(input_files, output_files=None):
    """Convert audio files to mp3"""
    # Convert to paths
    input_files = [Path(file) for file in input_files]

    # Default output filename is same as input but with MP3 extension
    if output_files is None:
        output_files = [file.with_suffix('.jpg') for file in input_files]

    # Multiprocess conversion
    with mp.Pool() as pool:
        pool.starmap(from_file_to_file, zip(input_files, output_files))

    # for input_file, output_file in zip(input_files, output_files):
    #     from_file_to_file(input_file, output_file)


###############################################################################
# Entry point
###############################################################################


def expand_files(files):
    """Expands a wildcard to a list of paths for Windows compatibility"""
    # Split at whitespace
    files = files.split()

    # Handle wildcard expansion
    if len(files) == 1 and '*' in files[0]:
        files = glob.glob(files[0])

    # Convert to Path objects
    return files


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(description='Convert images to JPEG')

    # Handle wildcards across platforms
    if os.name == 'nt':
        parser.add_argument(
            '--input_files',
            type=expand_files,
            help='The image files to convert to jpeg')
    else:
        parser.add_argument(
            '--input_files',
            nargs='+',
            help='The image files to convert to jpeg')

    parser.add_argument(
        '--output_files',
        type=Path,
        nargs='+',
        help='The corresponding output files. ' +
             'Uses same filename with jpg extension by default')
    return parser.parse_args()


if __name__ == '__main__':
    from_files_to_files(**vars(parse_args()))



import reseval


###############################################################################
# Evaluation file storage
###############################################################################


def create(config, directory, local=False):
    """Create file storage and upload files"""
    print('Creating storage...')

    # Make sure client public storage doesn't have old evaluation files
    reseval.storage.client.destroy(config)

    # Create and upload
    module(config, local).create(config, directory)


def destroy(name):
    """Destroy file storage"""
    local = reseval.is_local(name)
    config = reseval.load.config_by_name(name)
    module(config, local).destroy(config['name'])


def upload(name, file_or_directory):
    """Upload a file or directory to storage"""
    local = reseval.is_local(name)
    config = reseval.load.config_by_name(name)
    return module(config, local).upload(name, file_or_directory)


###############################################################################
# Utilities
###############################################################################


def module(config, local=False):
    """Get the storage module to use"""
    if local:
        return reseval.storage.client
    storage = config['storage']
    if storage == 'aws':
        return reseval.storage.aws
    raise ValueError(f'Storage method {storage} is not recognized')
